"""
Created on 29.09.2017

@author: lemmerfn
"""
import numbers
from collections import namedtuple
from functools import total_ordering

import numpy as np

import pysubgroup as ps


@total_ordering
class NumericTarget:
    statistic_types = (
        "size_sg",
        "size_dataset",
        "mean_sg",
        "mean_dataset",
        "std_sg",
        "std_dataset",
        "median_sg",
        "median_dataset",
        "max_sg",
        "max_dataset",
        "min_sg",
        "min_dataset",
        "mean_lift",
        "median_lift",
    )

    def __init__(self, target_variable):
        self.target_variable = target_variable

    def __repr__(self):
        return "T: " + str(self.target_variable)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__  # pragma: no cover

    def __lt__(self, other):
        return str(self) < str(other)  # pragma: no cover

    def get_attributes(self):
        return [self.target_variable]

    def get_base_statistics(self, subgroup, data):
        cover_arr, size_sg = ps.get_cover_array_and_size(subgroup, len(data), data)
        all_target_values = data[self.target_variable]
        sg_target_values = all_target_values[cover_arr]
        instances_dataset = len(data)
        instances_subgroup = size_sg
        mean_sg = np.mean(sg_target_values)
        mean_dataset = np.mean(all_target_values)
        return (instances_dataset, mean_dataset, instances_subgroup, mean_sg)

    def calculate_statistics(self, subgroup, data, cached_statistics=None):
        if cached_statistics is None or not isinstance(cached_statistics, dict):
            statistics = {}
        elif all(k in cached_statistics for k in NumericTarget.statistic_types):
            return cached_statistics
        else:
            statistics = cached_statistics

        cover_arr, _ = ps.get_cover_array_and_size(subgroup, len(data), data)
        all_target_values = data[self.target_variable].to_numpy()
        sg_target_values = all_target_values[cover_arr]

        statistics["size_sg"] = len(sg_target_values)
        statistics["size_dataset"] = len(data)
        statistics["mean_sg"] = np.mean(sg_target_values)
        statistics["mean_dataset"] = np.mean(all_target_values)
        statistics["std_sg"] = np.std(sg_target_values)
        statistics["std_dataset"] = np.std(all_target_values)
        statistics["median_sg"] = np.median(sg_target_values)
        statistics["median_dataset"] = np.median(all_target_values)
        statistics["max_sg"] = np.max(sg_target_values)
        statistics["max_dataset"] = np.max(all_target_values)
        statistics["min_sg"] = np.min(sg_target_values)
        statistics["min_dataset"] = np.min(all_target_values)
        statistics["mean_lift"] = statistics["mean_sg"] / statistics["mean_dataset"]
        statistics["median_lift"] = (
            statistics["median_sg"] / statistics["median_dataset"]
        )
        return statistics


def read_median(tpl):
    return tpl.median


def read_mean(tpl):
    return tpl.mean


def calc_sorted_median(arr):
    half = (len(arr) - 1) // 2
    if len(arr) % 2 == 0:
        return (arr[half] + arr[half + 1]) / 2
    else:
        return arr[half]


class StandardQFNumeric(ps.BoundedInterestingnessMeasure):
    tpl = namedtuple("StandardQFNumeric_parameters", ("size_sg", "mean", "estimate"))
    mean_tpl = tpl
    median_tpl = namedtuple(
        "StandardQFNumeric_median_parameters", ("size_sg", "median", "estimate")
    )

    @staticmethod
    def standard_qf_numeric(a, _, mean_dataset, instances_subgroup, mean_sg):
        return instances_subgroup**a * (mean_sg - mean_dataset)

    def __init__(self, a, invert=False, estimator="default", centroid="mean"):
        if not isinstance(a, numbers.Number):
            raise ValueError(f"a is not a number. Received a={a}")
        self.a = a
        self.invert = invert

        self.dataset_statistics = None
        self.all_target_values = None
        self.has_constant_statistics = False

        if centroid == "median":
            if estimator == "default":
                estimator = "max"
            assert estimator in (
                "max",
                "order",
            ), "For median only estimator = max or order are possible"
            self.required_stat_attrs = ("size_sg", "median")
            self.agg = np.median
            self.tpl = StandardQFNumeric.median_tpl
            self.read_centroid = read_median
        elif centroid == "sorted_median":
            if estimator == "default":
                estimator = "max"
            assert estimator in (
                "max",
                "order",
            ), "For median only estimator = max or order are possible"
            self.required_stat_attrs = ("size_sg", "median")
            self.agg = calc_sorted_median
            self.tpl = StandardQFNumeric.median_tpl
            self.read_centroid = read_median
        elif centroid == "mean":
            if estimator == "default":
                estimator = "sum"
            self.required_stat_attrs = ("size_sg", "mean")
            self.agg = np.mean
            self.tpl = StandardQFNumeric.mean_tpl
            self.read_centroid = read_mean
        else:
            raise ValueError(
                f"centroid was {centroid} which is not in (median, sorted_median, mean)"
            )

        if estimator == "sum":
            self.estimator = StandardQFNumeric.Summation_Estimator(self)
        elif estimator == "max":
            self.estimator = StandardQFNumeric.Max_Estimator(self)
        elif estimator == "average":
            self.estimator = StandardQFNumeric.Max_Estimator(self)
        elif estimator == "order":
            if centroid == "mean":
                self.estimator = StandardQFNumeric.MeanOrdering_Estimator(self)
            else:
                raise NotImplementedError(
                    "Order estimation is not implemented for median qf"
                )
        else:
            raise ValueError(
                "estimator is not one of the following: "
                + str(["sum", "average", "order"])
            )

    def calculate_constant_statistics(self, data, target):
        data = self.estimator.get_data(data, target)
        self.all_target_values = data[target.target_variable].to_numpy()
        target_centroid = self.agg(self.all_target_values)
        data_size = len(data)
        self.dataset_statistics = self.tpl(data_size, target_centroid, None)
        self.estimator.calculate_constant_statistics(data, target)
        self.has_constant_statistics = True

    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        dataset = self.dataset_statistics
        return StandardQFNumeric.standard_qf_numeric(
            self.a,
            dataset.size_sg,
            self.read_centroid(dataset),
            statistics.size_sg,
            self.read_centroid(statistics),
        )

    def calculate_statistics(self, subgroup, target, data, statistics=None):
        cover_arr, sg_size = ps.get_cover_array_and_size(
            subgroup, len(self.all_target_values), data
        )
        sg_centroid = 0
        sg_target_values = 0
        if sg_size > 0:
            sg_target_values = self.all_target_values[cover_arr]
            sg_centroid = self.agg(sg_target_values)
            estimate = self.estimator.get_estimate(
                subgroup, sg_size, sg_centroid, cover_arr, sg_target_values
            )
        else:
            estimate = float("-inf")
        return self.tpl(sg_size, sg_centroid, estimate)

    def optimistic_estimate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        return statistics.estimate

    class Summation_Estimator:
        def __init__(self, qf):
            self.qf = qf
            self.indices_greater_centroid = None
            self.target_values_greater_centroid = None

        def get_data(self, data, target):
            return data

        def calculate_constant_statistics(
            self, data, target
        ):  # pylint: disable=unused-argument
            self.indices_greater_centroid = (
                self.qf.all_target_values
                > self.qf.read_centroid(self.qf.dataset_statistics)
            )
            self.target_values_greater_centroid = (
                self.qf.all_target_values
            )  # [self.indices_greater_mean]

        def get_estimate(
            self, subgroup, sg_size, sg_centroid, cover_arr, _
        ):  # pylint: disable=unused-argument
            larger_than_centroid = self.target_values_greater_centroid[cover_arr][
                self.indices_greater_centroid[cover_arr]
            ]
            size_greater_centroid = len(larger_than_centroid)
            sum_greater_centroid = np.sum(larger_than_centroid)

            return sum_greater_centroid - size_greater_centroid * self.qf.read_centroid(
                self.qf.dataset_statistics
            )

    class Max_Estimator:
        def __init__(self, qf):
            self.qf = qf
            self.indices_greater_centroid = None
            self.target_values_greater_centroid = None

        def get_data(self, data, target):
            return data

        def calculate_constant_statistics(
            self, data, target
        ):  # pylint: disable=unused-argument
            self.indices_greater_centroid = (
                self.qf.all_target_values
                > self.qf.read_centroid(self.qf.dataset_statistics)
            )
            self.target_values_greater_centroid = self.qf.all_target_values

        def get_estimate(
            self, subgroup, sg_size, sg_centroid, cover_arr, _
        ):  # pylint: disable=unused-argument
            larger_than_centroid = self.target_values_greater_centroid[cover_arr][
                self.indices_greater_centroid[cover_arr]
            ]
            size_greater_centroid = len(larger_than_centroid)
            if size_greater_centroid == 0:
                return -np.inf
            max_greater_centroid = np.max(larger_than_centroid)

            return size_greater_centroid**self.qf.a * (
                max_greater_centroid - self.qf.read_centroid(self.qf.dataset_statistics)
            )

    class MeanOrdering_Estimator:
        def __init__(self, qf):
            self.qf = qf
            self.indices_greater_centroid = None
            self._get_estimate = self.get_estimate_numpy
            self.use_numba = True
            self.numba_in_place = False

        def get_data(self, data, target):
            data.sort_values(target.get_attributes()[0], ascending=False, inplace=True)
            return data

        def calculate_constant_statistics(self, data, target):
            if not self.use_numba or self.numba_in_place:
                return
            try:
                from numba import njit  # pylint: disable=import-outside-toplevel

                # print('StandardQf_Numeric: Using numba for speedup')
            except ImportError:  # pragma: no cover
                return

            @njit
            def estimate_numba(values_sg, a, mean_dataset):  # pragma: no cover
                n = 1
                sum_values = 0
                max_value = -(10**10)
                for val in values_sg:
                    sum_values += val
                    mean_sg = sum_values / n
                    quality = n**a * (mean_sg - mean_dataset)
                    if quality > max_value:
                        max_value = quality
                    n += 1
                return max_value

            self._get_estimate = estimate_numba
            self.numba_in_place = True

        def get_estimate(
            self, subgroup, sg_size, sg_mean, cover_arr, target_values_sg
        ):  # pylint: disable=unused-argument
            if self.numba_in_place:
                return self._get_estimate(
                    target_values_sg, self.qf.a, self.qf.dataset_statistics.mean
                )
            else:
                return self._get_estimate(
                    target_values_sg, self.qf.a, self.qf.dataset_statistics.mean
                )

        def get_estimate_numpy(self, values_sg, _, mean_dataset):
            target_values_cs = np.cumsum(values_sg)
            sizes = np.arange(1, len(target_values_cs) + 1)
            mean_values = target_values_cs / sizes
            stats = StandardQFNumeric.mean_tpl(sizes, mean_values, mean_dataset)
            qualities = self.qf.evaluate(None, None, None, stats)
            optimistic_estimate = np.max(qualities)
            return optimistic_estimate


class StandardQFNumericMedian(ps.BoundedInterestingnessMeasure):
    tpl = namedtuple(
        "StandardQFNumericMedian_parameters",
        (
            "size_sg",
            "median",
            "estimate",
        ),  # this is here to allow older pickles to be loaded
    )

    def __init__(
        self,
    ):
        raise NotImplementedError(
            "StandardQFNumericMedian is no longer supported use "
            "StandardQFNumeric(centroid='median' instead)"
        )  # pragma: no cover


class StandardQFNumericTscore(ps.BoundedInterestingnessMeasure):
    tpl = namedtuple(
        "StandardQFNumericTscore_parameters", ("size_sg", "mean", "std", "estimate")
    )

    @staticmethod
    def t_score(mean_dataset, instances_subgroup, mean_sg, std_sg):
        if std_sg == 0:
            return 0
        else:
            return (instances_subgroup**0.5 * (mean_sg - mean_dataset)) / std_sg

    def __init__(self, invert=False):
        self.invert = invert
        self.required_stat_attrs = ("size_sg", "mean", "std")
        self.dataset_statistics = None
        self.all_target_values = None
        self.has_constant_statistics = False

    def calculate_constant_statistics(self, data, target):
        self.all_target_values = data[target.target_variable].to_numpy()
        target_mean = np.mean(self.all_target_values)
        target_std = np.std(self.all_target_values)
        data_size = len(data)
        self.dataset_statistics = StandardQFNumericTscore.tpl(
            data_size, target_mean, target_std, np.inf
        )
        self.has_constant_statistics = True

    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        dataset = self.dataset_statistics
        return StandardQFNumericTscore.t_score(
            dataset.mean,
            statistics.size_sg,
            statistics.mean,
            statistics.std,
        )

    def calculate_statistics(self, subgroup, target, data, statistics=None):
        cover_arr, sg_size = ps.get_cover_array_and_size(
            subgroup, len(self.all_target_values), data
        )
        sg_mean = np.array([0])
        sg_std = np.array([0])
        sg_target_values = 0
        if sg_size > 0:
            sg_target_values = self.all_target_values[cover_arr]
            sg_mean = np.mean(sg_target_values)
            sg_std = np.std(sg_target_values)
            estimate = np.inf
        else:
            estimate = float("-inf")
        return StandardQFNumericTscore.tpl(sg_size, sg_mean, sg_std, estimate)

    def optimistic_estimate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        return statistics.estimate


# TODO Update to new format
# class GAStandardQFNumeric(ps.AbstractInterestingnessMeasure):
#    def __init__(self, a, invert=False):
#        self.a = a
#        self.invert = invert
#
#    def evaluate_from_dataset(self, data, subgroup, weighting_attribute=None):
#        (instances_dataset, _, instances_subgroup, mean_sg) = \
#           subgroup.get_base_statistics(data, weighting_attribute)
#        if instances_subgroup in (0, instances_dataset):
#            return 0
#        max_mean = get_max_generalization_mean(data, subgroup, weighting_attribute)
#        relative_size = (instances_subgroup / instances_dataset)
#        return ps.conditional_invert(
#           relative_size ** self.a * (mean_sg - max_mean), self.invert)

#    def supports_weights(self):
#        return True

#    def is_applicable(self, subgroup):
#        return isinstance(subgroup.target, NumericTarget)


# def get_max_generalization_mean(data, subgroup, weighting_attribute=None):
#    selectors = subgroup.subgroup_description.selectors
#    generalizations = ps.powerset(selectors)
#    max_mean = 0
#    for sels in generalizations:
#        sg = ps.Subgroup(subgroup.target, ps.Conjunction(list(sels)))
#        mean_sg = sg.get_base_statistics(data, weighting_attribute)[3]
#        max_mean = max(max_mean, mean_sg)
#    return max_mean



from io import StringIO

import pandas as pd
import pkg_resources
from scipy.io import arff


def get_credit_data():
    s_io = StringIO(
        str(pkg_resources.resource_string("pysubgroup", "data/credit-g.arff"), "utf-8")
    )
    return pd.DataFrame(arff.loadarff(s_io)[0])


def get_titanic_data():
    s_io = StringIO(
        str(pkg_resources.resource_string("pysubgroup", "data/titanic.csv"), "utf-8")
    )
    return pd.read_csv(s_io, sep="\t", header=[0])



import unittest

import pandas as pd

import pysubgroup as ps


class TestBinaryTarget(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame.from_records(
            [
                (1, 1, 0),
                (
                    1,
                    1,
                    1,
                ),
                (1, 1, 0),
                (1, 0, 1),
            ],
            columns=("A", "B", "C"),
        )
        self.selector = ps.EqualitySelector("A", 1)

    def test_eq(self):
        T1 = ps.BinaryTarget("A", 1)
        T2 = ps.BinaryTarget("A", 1)
        self.assertTrue(T1 == T2)

    def test_lt(self):
        T1 = ps.BinaryTarget("A", 1)
        T2 = ps.BinaryTarget("B", 1)
        self.assertTrue(T1 < T2)

    def test_init_errors(self):
        with self.assertRaises(ValueError):
            ps.BinaryTarget()
        with self.assertRaises(ValueError):
            ps.BinaryTarget("A", 1, self.selector)

    def test_calculate_statistics_for_cached(self):
        target = ps.BinaryTarget("C", 1)
        statistics = target.calculate_statistics(self.selector, self.df)
        statistics2 = target.calculate_statistics(self.selector, self.df, statistics)
        self.assertIs(statistics, statistics2)

    def test_LiftQf(self):
        qf = ps.LiftQF()
        self.assertIsInstance(qf, ps.StandardQF)

    def test_SimpleBinomialQF(self):
        qf = ps.SimpleBinomialQF()
        self.assertIsInstance(qf, ps.StandardQF)

    def test_LWRAccQF(self):
        qf = ps.WRAccQF()
        self.assertIsInstance(qf, ps.StandardQF)


if __name__ == "__main__":
    unittest.main()



import unittest

import pandas as pd
from t_utils import assertResultEqual

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestGpGrowth(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame.from_records(
            [
                (1, 1, 0),
                (
                    1,
                    1,
                    1,
                ),
                (1, 1, 0),
                (1, 0, 1),
            ],
            columns=("A", "B", "C"),
        )

        target = ps.FITarget()
        QF = ps.CountQF()

        self.df1 = pd.DataFrame.from_records(
            [
                (1, 1, 1, 0),
                (1, 1, 0, 1),
                (1, 1, 1, 0),
                (1, 1, 0, 1),
                (1, 1, 0, 0),
                (1, 0, 0, 0),
            ],
            columns=("A", "B", "C", "D"),
        )
        selectors1 = [ps.EqualitySelector(x, 1) for x in self.df1.columns]
        self.task1 = ps.SubgroupDiscoveryTask(
            self.df1, target, selectors1, result_set_size=20, depth=4, qf=QF
        )
        self.solution1 = """6 A==1
        6 Dataset
        5 B==1
        5 A==1 AND B==1
        2 D==1
        2 A==1 AND D==1
        2 A==1 AND B==1 AND D==1
        2 A==1 AND B==1 AND C==1
        2 B==1 AND D==1
        2 A==1 AND C==1
        2 B==1 AND C==1
        2 C==1"""

        self.df2 = pd.DataFrame.from_records(
            [
                (1, 1, 1, 0),
                (1, 1, 0, 1, 1),
                (1, 1, 1, 0),
                (1, 1, 0, 1, 1),
            ],
            columns=("A", "B", "C", "D", "E"),
        )
        selectors2 = [ps.EqualitySelector(x, 1) for x in self.df2.columns]
        self.task2 = ps.SubgroupDiscoveryTask(
            self.df2, target, selectors2, result_set_size=20, depth=4, qf=QF
        )

        self.df3 = pd.DataFrame.from_records(
            [
                (
                    1,
                    1,
                ),
                (0, 1),
                (1, 0),
                (1, 0),
            ],
            columns=("A", "B"),
        )
        selectors3 = [ps.EqualitySelector(x, 1) for x in self.df3.columns]
        self.task3 = ps.SubgroupDiscoveryTask(
            self.df3, target, selectors3, result_set_size=20, depth=4, qf=QF
        )
        self.all_selectors = {x: ps.EqualitySelector(x, 1) for x in self.df2.columns}

        self.task4 = ps.SubgroupDiscoveryTask(
            self.df1,
            target,
            selectors1,
            result_set_size=20,
            depth=4,
            qf=QF,
            constraints=[ps.MinSupportConstraint(3)],
        )
        self.solution4 = """6 A==1
        6 Dataset
        5 B==1
        5 A==1 AND B==1"""

        self.task5 = ps.SubgroupDiscoveryTask(
            self.df1,
            target,
            selectors1,
            result_set_size=20,
            depth=4,
            qf=QF,
            constraints=[ps.MinSupportConstraint(10)],
        )
        self.solution5 = """"""

    def test_export_fi(self):
        target = ps.FITarget()
        searchspace = ps.create_selectors(self.df1)
        task = ps.SubgroupDiscoveryTask(
            self.df1, target, searchspace, result_set_size=5, depth=2, qf=ps.CountQF()
        )
        ps.GpGrowth().to_file(task, "./test_gp_fi.txt")

    def test_export_binary(self):
        target = ps.BinaryTarget("A", 1)
        searchspace = ps.create_selectors(self.df, ignore=["A"])
        task = ps.SubgroupDiscoveryTask(
            self.df,
            target,
            searchspace,
            result_set_size=5,
            depth=2,
            qf=ps.StandardQF(0.5),
        )
        ps.GpGrowth().to_file(task, "./test_gp_binary.txt")

    def test_export_model(self):
        model = ps.PolyRegression_ModelClass("A", "B")
        QF = ps.EMM_Likelihood(model)
        searchspace = ps.create_selectors(self.df)
        task = ps.SubgroupDiscoveryTask(
            self.df, None, searchspace, result_set_size=5, depth=2, qf=QF
        )
        ps.GpGrowth().to_file(task, "./test_gp_model.txt")

    def test_gp_modes_restricted(self):
        with self.assertRaises(AssertionError):
            ps.GpGrowth(mode="fail").execute(self.task1)

    def test_gp_qf_requires_cov(self):
        tmp = self.task1.qf.gp_requires_cover_arr
        # del self.task1.qf.gp_requires_cover_arr
        self.task1.qf.gp_requires_cover_arr = True
        with self.assertWarns(UserWarning):
            result = ps.GpGrowth(mode="t_d").execute(self.task1)
        self.task1.qf.gp_requires_cover_arr = tmp
        assertResultEqual(self, result, self.solution1)

    def test_gp_b_u_simple1(self):
        result = ps.GpGrowth(mode="b_u").execute(self.task1)
        # print_result(result)
        assertResultEqual(self, result, self.solution1)

    def test_gp_simple1(self):
        with self.assertWarns(UserWarning):
            result = ps.GpGrowth(mode="t_d").execute(self.task1)
        # print_result(result)
        assertResultEqual(self, result, self.solution1)

        # self.assertEqual(results[0], (6, ps.Conjunction([self.all_selectors["A"]]) ))

    def test_gp_simple2(self):
        with self.assertWarns(UserWarning):
            result = ps.GpGrowth(mode="t_d").execute(self.task2)
        # print_result(result)
        assertResultEqual(
            self,
            result,
            """ 4 A==1 AND B==1
            4 A==1
            4 Dataset
            4 B==1
            2 A==1 AND B==1 AND E==1
            2 A==1 AND B==1 AND D==1 AND E==1
            2 B==1 AND E==1
            2 A==1 AND B==1 AND D==1
            2 A==1 AND E==1
            2 A==1 AND D==1 AND E==1
            2 B==1 AND D==1
            2 B==1 AND D==1 AND E==1
            2 A==1 AND B==1 AND D==1
            2 C==1
            2 A==1 AND C==1
            2 A==1 AND D==1
            2 A==1 AND D==1
            2 B==1 AND C==1
            2 B==1 AND D==1
            2 A==1 AND B==1 AND C==1""",
        )

    def test_gp_simple3(self):
        with self.assertWarns(UserWarning):
            result = ps.GpGrowth(mode="t_d").execute(self.task3)
        # print_result(result)
        assertResultEqual(
            self,
            result,
            """4 Dataset
            3 A==1
            2 B==1
            1 A==1 AND B==1
            1 B==1""",
        )

    def test_gp_simple4(self):
        result = ps.GpGrowth(mode="t_d").execute(self.task4)
        # print_result(result)
        assertResultEqual(self, result, self.solution4)

    def test_gp_simple5(self):
        result = ps.GpGrowth(mode="t_d").execute(self.task5)
        # print_result(result)
        assertResultEqual(self, result, self.solution5)

    def test_gp_credit(self):
        data = get_credit_data()
        # warnings.filterwarnings("error")
        searchSpace_Nominal = ps.create_nominal_selectors(
            data, ignore=["duration", "credit_amount", "class"]
        )
        searchSpace_Numeric = ps.create_numeric_selectors(
            data, ignore=["duration", "credit_amount", "class"]
        )
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        # searchSpace = [
        #   ps.EqualitySelector("checking_status", b'<0'),
        #   ps.EqualitySelector("own_telephone", b'yes'),
        #   ps.IntervalSelector("age", 30, 36)]
        # QF=model_target.EMM_Likelihood(
        #   model_target.PolyRegression_ModelClass(
        #       x_name='duration', y_name='credit_amount'))
        # target = ps.FITarget()
        # QF=ps.CountQF()

        QF = ps.StandardQF(0.5)
        target = ps.BinaryTarget("class", b"bad")
        task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=30,
            depth=2,
            qf=QF,
            constraints=[ps.MinSupportConstraint(30)],
        )
        result = ps.GpGrowth(mode="t_d").execute(task)
        # print_result(result)
        assertResultEqual(
            self,
            result,
            """0.10866138825404954 checking_status=='b'<0'' AND foreign_worker=='b'yes''
            0.10705790652183149 checking_status=='b'<0'' AND job=='b'skilled''
            0.10371988780388462 checking_status=='b'<0'' AND other_parties=='b'none''
            0.1032107831257212 checking_status=='b'<0'' AND savings_status=='b'<100''
            0.10086921502691486 checking_status=='b'<0''
            0.10060089731478809 checking_status=='b'<0'' AND num_dependents==1.0
            0.09924929488520949 checking_status=='b'<0'' AND installment_commitment==4.0
            0.09469906064957609 checking_status=='b'<0'' AND existing_credits==1.0
            0.08675031975461583 checking_status=='b'<0'' AND own_telephone=='b'none''
            0.08528028654224418 checking_status=='b'<0'' AND other_payment_plans=='b'none''
            0.08499999999999999 checking_status=='b'<0'' AND credit_history=='b'existing paid''
            0.07816842732397161 checking_status=='b'<0'' AND residence_since==4.0
            0.07334123103829637 age: [30.0:36.0[ AND checking_status=='b'<0''
            0.07278624758728697 checking_status=='b'<0'' AND property_magnitude=='b'car''
            0.07155417527999328 checking_status=='b'<0'' AND property_magnitude=='b'no known property''
            0.06928203230275509 credit_history=='b'no credits/all paid'' AND savings_status=='b'<100''
            0.06734716791106383 credit_history=='b'no credits/all paid'' AND other_parties=='b'none''
            0.06659868328566726 checking_status=='b'<0'' AND purpose=='b'new car''
            0.06607158652139772 checking_status=='b'<0'' AND personal_status=='b'female div/dep/mar''
            0.0654846187598099 checking_status=='b'<0'' AND housing=='b'own''
            0.065 credit_history=='b'no credits/all paid''
            0.06471832459560074 checking_status=='b'<0'' AND housing=='b'rent''
            0.0646366361813647 credit_history=='b'no credits/all paid'' AND foreign_worker=='b'yes''
            0.06411591895800751 checking_status=='b'0<=X<200'' AND property_magnitude=='b'no known property''
            0.06399448505650358 other_payment_plans=='b'bank'' AND purpose=='b'new car''
            0.06394465512393448 checking_status=='b'<0'' AND personal_status=='b'male single''
            0.063727937358744 age<26.0 AND purpose=='b'new car''
            0.06365809407433175 age<26.0 AND checking_status=='b'<0''
            0.06301438731059822 age<26.0 AND savings_status=='b'<100''
            0.06272082938724895 credit_history=='b'all paid'' AND other_parties=='b'none''""",  # noqa: E501
        )


if __name__ == "__main__":
    unittest.main()



import pandas as pd

import pysubgroup as ps

data = pd.read_table("../data/titanic.csv")
target = ps.BinaryTarget("Survived", 0)
search_space = ps.create_selectors(data, ignore=["Survived"])
task = ps.SubgroupDiscoveryTask(
    data,
    target,
    search_space,
    result_set_size=5,
    depth=2,
    qf=ps.CombinedInterestingnessMeasure(
        [ps.StandardQF(1), ps.GeneralizationAware_StandardQF(1)]
    ),
)

result = ps.SimpleDFS().execute(task, use_optimistic_estimates=False)

print(result.to_dataframe())



import pysubgroup as ps
from pysubgroup.datasets import get_credit_data

data = get_credit_data()

target = ps.NumericTarget("credit_amount")
sgd = ps.EqualitySelector("purpose", b"other")

stats = target.calculate_statistics(sgd, data)
print(stats)


qf = ps.StandardQFNumeric(1.0)
score = qf.evaluate(sgd, target, data)
print(score)

score = qf.evaluate(sgd, target, data, stats)
print(score)



import os
import pickle
import tempfile
import unittest

import numpy as np
import pandas as pd

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestRelationsMethods(unittest.TestCase):
    def test_pickle(self):
        A1 = ps.EqualitySelector("A", 1)
        with tempfile.TemporaryDirectory() as td:
            f_name = os.path.join(td, "test.pickle")

            with open(f_name, "wb") as f:
                pickle.dump(A1, f)

            with open(f_name, "rb") as f:
                A2 = pickle.load(f)

            assert A1 == A2


class TestBasics(unittest.TestCase):
    def test_get_cover_array_and_size(self):
        df = get_credit_data()  #
        # len(df)==1000
        self.assertEqual(ps.get_cover_array_and_size(slice(None), data=df)[1], len(df))
        self.assertEqual(ps.get_cover_array_and_size(slice(3, 11), data=df)[1], 8)
        self.assertEqual(ps.get_cover_array_and_size(slice(900, 1100), data=df)[1], 100)
        with self.assertRaises(ValueError):
            ps.get_cover_array_and_size(slice(900, 1100))
        self.assertEqual(
            ps.get_cover_array_and_size(np.array([1, 0, 1], dtype=bool))[1], 2
        )
        with self.assertRaises(NotImplementedError):
            ps.get_cover_array_and_size(np.array(["s", "b"]))

    def test_get_size(self):
        df = get_credit_data()  #
        # len(df)==1000
        self.assertEqual(ps.get_size(slice(None), data_len=len(df)), len(df))
        self.assertEqual(ps.get_size(slice(None), data=df), len(df))
        self.assertEqual(ps.get_size(slice(3, 11), data=df), 8)
        self.assertEqual(ps.get_size(slice(900, 1100), data=df), 100)
        with self.assertRaises(ValueError):
            ps.get_size(slice(900, 1100))
        self.assertEqual(ps.get_size(np.array([1, 0, 1], dtype=bool)), 2)
        self.assertEqual(
            ps.get_size(ps.EqualitySelector("checking_status", b"<0"), data=df), 274
        )
        self.assertEqual(ps.get_size(np.array([1, 2, 3], dtype=np.int32)), 3)
        with self.assertRaises(NotImplementedError):
            ps.get_size(np.array(["s", "b"]))
        with self.assertRaises(ValueError):
            ps.get_size(slice(None))

        self.assertTrue(ps.get_size(slice(None), data=df), 152)

    def test_EqualitySelector(self):
        with self.assertRaises(TypeError):
            ps.EqualitySelector(None, 1)
        with self.assertRaises(TypeError):
            ps.EqualitySelector("a", None)
        with self.assertRaises(TypeError):
            ps.EqualitySelector(None, None)
        sel = ps.EqualitySelector("A", np.nan)

        self.assertEqual(sel.selectors, (sel,))
        self.assertFalse(sel is None)

    def test_NegatedSelector(self):
        df = pd.DataFrame.from_records(
            [
                (1, 1, 0),
                (
                    1,
                    1,
                    1,
                ),
                (1, 1, 0),
                (1, 0, 1),
            ],
            columns=("A", "B", "C"),
        )
        A_sel = ps.EqualitySelector("A", 1)

        A_neg = ps.NegatedSelector(A_sel)
        self.assertEqual(A_neg.attribute_name, "A")
        self.assertEqual(A_neg.selectors, (A_neg,))
        np.testing.assert_array_equal(A_neg.covers(df), [0, 0, 0, 0])

        B_sel = ps.EqualitySelector("B", 1)
        B_neg = ps.NegatedSelector(B_sel)

        np.testing.assert_array_equal(B_neg.covers(df), [0, 0, 0, 1])

    def test_IntervalSelector(self):
        sel = ps.IntervalSelector("A", float("-inf"), float("+inf"))
        self.assertEqual(str(sel), "A = anything")
        self.assertEqual(sel.selectors, (sel,))

        sel = ps.IntervalSelector("A", float("-inf"), 0)
        sel = ps.IntervalSelector("A", 0, float("+inf"))
        sel = ps.IntervalSelector("A", 0, 1)

        sel = ps.IntervalSelector("age", 60, float("+inf"), "the_elderly")
        self.assertEqual(str(sel), "the_elderly")

        with self.assertRaises(AssertionError):
            ps.IntervalSelector("age", 70, 60)

    def test_remove_target_attributes(self):
        selectors = [
            ps.IntervalSelector("A", 1, 2),
            ps.EqualitySelector("B", 1),
            ps.EqualitySelector("C", 1),
        ]

        fi_target = ps.FITarget()
        fi_sel = ps.remove_target_attributes(selectors, fi_target)
        self.assertEqual(selectors[:], fi_sel)

        num_target = ps.NumericTarget("C")
        num_sel = ps.remove_target_attributes(selectors, num_target)
        self.assertEqual(selectors[:-1], num_sel)

        bin_target = ps.BinaryTarget("A", -10)
        bin_sel = ps.remove_target_attributes(selectors, bin_target)
        self.assertEqual(selectors[1:], bin_sel)

    def test_EqualitySelector_from_str(self):
        selector1 = ps.EqualitySelector.from_str("A=='hello'")
        selector2 = ps.EqualitySelector("A", "hello")
        self.assertEqual(selector1, selector2)
        self.assertIs(selector1, selector2)

    def test_IntervalSelector_from_str(self):
        def test_str(s, attribute_name, lb, ub):
            selector = ps.IntervalSelector.from_str(s)
            _selector = ps.IntervalSelector(attribute_name, lb, ub)
            self.assertEqual(selector, _selector)
            self.assertIs(selector, _selector)

        test_str("A = anything", "A", float("-inf"), float("+inf"))
        test_str("A >= 10", "A", 10, float("+inf"))
        test_str("A >= 10.0", "A", 10.0, float("+inf"))
        test_str("A < 12.3", "A", float("-inf"), 12.3)
        test_str("A < 12", "A", float("-inf"), 12)
        test_str("A : [12:34[", "A", 12, 34)
        test_str("A : [12.3:45.6[", "A", 12.3, 45.6)

        with self.assertRaises(ValueError):
            ps.IntervalSelector.from_str("A == B")

    def test_create_nominal_selectors_for_attribut(self):
        df = pd.DataFrame({"A": np.array([1, 1, 0], dtype=bool)})
        selectors = ps.create_nominal_selectors_for_attribute(df, "A")
        self.assertTrue(all(sel.is_bool for sel in selectors))

    def test_Conjunction(self):
        conj = ps.Conjunction(
            [ps.EqualitySelector.from_str("A==0"), ps.EqualitySelector.from_str("B==1")]
        )
        with self.assertRaises(RuntimeError):
            conj.append_or(ps.EqualitySelector.from_str("C==12"))
        with self.assertRaises(RuntimeError):
            conj.pop_or()
        sel = conj.pop_and()
        self.assertIs(sel, ps.EqualitySelector("B", 1))
        self.assertEqual(conj, ps.Conjunction([ps.EqualitySelector("A", 0)]))

    def test_Disjunction(self):
        from copy import copy  # pylint: disable=import-outside-toplevel

        dis = ps.Disjunction(
            [ps.EqualitySelector.from_str("A==0"), ps.EqualitySelector.from_str("B==1")]
        )
        self.assertEqual(len(dis), 2)
        dis2 = copy(dis)
        self.assertEqual(dis, dis2)
        self.assertFalse(dis is dis2)
        with self.assertRaises(RuntimeError):
            dis.append_and(ps.EqualitySelector.from_str("C==12"))

        df = pd.DataFrame({"A": np.array([1, 1, 0], dtype=bool)})
        dis = ps.Disjunction([])
        np.testing.assert_array_equal(dis.covers(df), [False, False, False])

    def test_create_numeric_selectors_for_attribute(self):
        from t_utils import (
            interval_selectors_from_str,  # pylint:disable=import-outside-toplevel
        )

        df = pd.DataFrame({"A": np.array([1, 1, 0, 3, 3, 4, 5], dtype=int)})
        selectors = ps.create_numeric_selectors_for_attribute(
            df, "A", nbins=4, intervals_only=False
        )
        result_selectors = interval_selectors_from_str(
            """A>=1
            A<1
            A>=3
            A<3
            A>=4
            A<4"""
        )
        self.assertEqual(selectors, result_selectors)


if __name__ == "__main__":
    unittest.main()



"""
Created on 02.05.2016

@author: lemmerfn
"""
import itertools
from collections.abc import Iterable
from functools import partial
from heapq import heappop, heappush

import numpy as np

import pysubgroup as ps

from .algorithms import SubgroupDiscoveryTask


def minimum_required_quality(result, task):
    if len(result) < task.result_set_size:
        return task.min_quality
    else:
        return result[0][0]


def prepare_subgroup_discovery_result(result, task):
    result_filtered = [tpl for tpl in result if tpl[0] > task.min_quality]
    result_filtered.sort(key=lambda x: x[0], reverse=True)
    result_filtered = result_filtered[: task.result_set_size]
    return result_filtered


# Returns the cutpoints for discretization
def equal_frequency_discretization(
    data, attribute_name, nbins=5, weighting_attribute=None
):
    cutpoints = []
    if weighting_attribute is None:
        cleaned_data = data[attribute_name]
        cleaned_data = cleaned_data[~np.isnan(cleaned_data)]
        sorted_data = sorted(cleaned_data)
        number_instances = len(sorted_data)
        for i in range(1, nbins):
            position = i * number_instances // nbins
            while True:
                if position >= number_instances:
                    break
                val = sorted_data[position]
                if val not in cutpoints:
                    break
                position += 1
            # print (sorted_data [position])
            if val not in cutpoints:
                cutpoints.append(val)
    else:
        cleaned_data = data[[attribute_name, weighting_attribute]]
        cleaned_data = cleaned_data[~np.isnan(cleaned_data[attribute_name])]
        cleaned_data.sort(order=attribute_name)

        overall_weights = cleaned_data[weighting_attribute].sum()
        remaining_weights = overall_weights
        bin_size = overall_weights / nbins
        sum_of_weights = 0
        for row in cleaned_data:
            sum_of_weights += row[weighting_attribute]
            if sum_of_weights > bin_size:
                if not row[attribute_name] in cutpoints:
                    cutpoints.append(row[attribute_name])
                    remaining_weights = remaining_weights - sum_of_weights
                    if remaining_weights < 1.5 * (bin_size):
                        break
                    sum_of_weights = 0
    return cutpoints


def conditional_invert(val, invert):
    return -2 * (invert - 0.5) * val


def results_df_autoround(df):
    return df.round(
        {
            "quality": 3,
            "size_sg": 0,
            "size_dataset": 0,
            "positives_sg": 0,
            "positives_dataset": 0,
            "size_complement": 0,
            "relative_size_sg": 3,
            "relative_size_complement": 3,
            "coverage_sg": 3,
            "coverage_complement": 3,
            "target_share_sg": 3,
            "target_share_complement": 3,
            "target_share_dataset": 3,
            "lift": 3,
            "size_sg_weighted": 1,
            "size_dataset_weighted": 1,
            "positives_sg_weighted": 1,
            "positives_dataset_weighted": 1,
            "size_complement_weighted": 1,
            "relative_size_sg_weighted": 3,
            "relative_size_complement_weighted": 3,
            "coverage_sg_weighted": 3,
            "coverage_complement_weighted": 3,
            "target_share_sg_weighted": 3,
            "target_share_complement_weighted": 3,
            "target_share_dataset_weighted": 3,
            "lift_weighted": 3,
        }
    )


def perc_formatter(x):
    return "{0:.1f}%".format(x * 100)


def float_formatter(x, digits=2):
    return ("{0:." + str(digits) + "f}").format(x)


def is_categorical_attribute(data, attribute_name):
    return attribute_name in data.select_dtypes(exclude=["number"]).columns.values


def is_numerical_attribute(data, attribute_name):
    return attribute_name in data.select_dtypes(include=["number"]).columns.values


def remove_selectors_with_attributes(selector_list, attribute_list):
    return [x for x in selector_list if x.attributeName not in attribute_list]


def derive_effective_sample_size(weights):
    return sum(weights) ** 2 / sum(weights**2)


# from https://docs.python.org/3/library/itertools.html#recipes
def powerset(iterable, max_length=None):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    if max_length is None:
        max_length = len(s)
    if max_length < len(s):
        max_length = len(s)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(max_length)
    )


def overlap(sg, another_sg, data):
    cover_sg = sg.covers(data)
    cover_another_sg = another_sg.covers(data)
    union = np.logical_or(cover_sg, cover_another_sg)
    intercept = np.logical_and(cover_sg, cover_another_sg)
    sim = np.sum(intercept) / np.sum(union)
    return sim


#####
# bitset operations
#####
def to_bits(list_of_ints):
    v = 0
    for x in list_of_ints:
        v += 1 << x
    return v


def count_bits(bitset_as_int):
    c = 0
    while bitset_as_int > 0:
        c += 1
        bitset_as_int &= bitset_as_int - 1
    return c


def find_set_bits(bitset_as_int):
    while bitset_as_int > 0:
        x = bitset_as_int.bit_length() - 1
        yield x
        bitset_as_int = bitset_as_int - (1 << x)


#####
# TID-list operations
#####
def intersect_of_ordered_list(list_1, list_2):
    result = []
    i = 0
    j = 0
    while i < len(list_1) and j < len(list_2):
        if list_1[i] < list_2[j]:
            i += 1
        elif list_2[j] < list_1[i]:
            j += 1
        else:
            result.append(list_1[i])
            j += 1
            i += 1
    return result


class BaseTarget:
    def all_statistics_present(self, cached_statistics):
        if isinstance(cached_statistics, dict) and all(
            expected_value in cached_statistics
            for expected_value in self.__class__.statistic_types
        ):  # pylint: disable=no-member
            return True
        return False


class SubgroupDiscoveryResult:
    def __init__(self, results, task):
        self.task = task
        self.results = results
        assert isinstance(results, Iterable)

    def to_descriptions(self, include_stats=False):
        if include_stats:
            return list(self.results)
        else:
            return [(qual, sgd) for qual, sgd, stats in self.results]

    def to_table(
        self, statistics_to_show=None, print_header=True, include_target=False
    ):
        if statistics_to_show is None:
            statistics_to_show = type(self.task.target).statistic_types
        table = []
        if print_header:
            row = ["quality", "subgroup"]
            if include_target:
                row.append("target")
            for stat in statistics_to_show:
                row.append(stat)
            table.append(row)
        for q, sg, stats in self.results:
            stats = self.task.target.calculate_statistics(sg, self.task.data, stats)
            row = [q, sg]
            if include_target:
                row.append(self.task.target)
            for stat in statistics_to_show:
                row.append(stats[stat])
            table.append(row)
        return table

    def to_dataframe(
        self, statistics_to_show=None, autoround=False, include_target=False
    ):
        import pandas as pd  # pylint: disable=import-outside-toplevel

        if statistics_to_show is None:
            statistics_to_show = type(self.task.target).statistic_types
        res = self.to_table(statistics_to_show, True, include_target)
        headers = res.pop(0)
        df = pd.DataFrame(res, columns=headers)
        if autoround:
            df = results_df_autoround(df)
        return df

    def to_latex(self, statistics_to_show=None):
        if statistics_to_show is None:
            statistics_to_show = type(self.task.target).statistic_types
        df = self.to_dataframe(statistics_to_show)
        latex = df.to_latex(
            index=False,
            col_space=10,
            formatters={
                "quality": partial(float_formatter, digits=3),
                "size_sg": partial(float_formatter, digits=0),
                "size_dataset": partial(float_formatter, digits=0),
                "positives_sg": partial(float_formatter, digits=0),
                "positives_dataset": partial(float_formatter, digits=0),
                "size_complement": partial(float_formatter, digits=0),
                "relative_size_sg": perc_formatter,
                "relative_size_complement": perc_formatter,
                "coverage_sg": perc_formatter,
                "coverage_complement": perc_formatter,
                "target_share_sg": perc_formatter,
                "target_share_complement": perc_formatter,
                "target_share_dataset": perc_formatter,
                "lift": partial(float_formatter, digits=1),
            },
        )
        latex = latex.replace(" AND ", r" $\wedge$ ")
        return latex


def add_if_required(
    result,
    sg,
    quality,
    task: SubgroupDiscoveryTask,
    check_for_duplicates=False,
    statistics=None,
    explicit_result_set_size=None,
):
    """
    IMPORTANT:
        Only add/remove subgroups from `result` by using `heappop` and `heappush`
        to ensure order of subgroups by quality.
    """
    if explicit_result_set_size is None:
        explicit_result_set_size = task.result_set_size

    if quality >= task.min_quality:
        if not ps.constraints_satisfied(task.constraints, sg, statistics, task.data):
            return
        if check_for_duplicates and (quality, sg, statistics) in result:
            return
        if len(result) < explicit_result_set_size:
            heappush(result, (quality, sg, statistics))
        elif quality > result[0][0]:  # better than worst subgroup
            heappop(result)
            heappush(result, (quality, sg, statistics))



"""
Created on 29.09.2017

@author: lemmerfn
"""
from collections import namedtuple
from functools import total_ordering

import numpy as np

from pysubgroup.measures import (
    AbstractInterestingnessMeasure,
    BoundedInterestingnessMeasure,
    GeneralizationAwareQF_stats,
)

from .subgroup_description import EqualitySelector, get_cover_array_and_size
from .utils import BaseTarget, derive_effective_sample_size


@total_ordering
class BinaryTarget(BaseTarget):
    statistic_types = (
        "size_sg",
        "size_dataset",
        "positives_sg",
        "positives_dataset",
        "size_complement",
        "relative_size_sg",
        "relative_size_complement",
        "coverage_sg",
        "coverage_complement",
        "target_share_sg",
        "target_share_complement",
        "target_share_dataset",
        "lift",
    )

    def __init__(self, target_attribute=None, target_value=None, target_selector=None):
        """
        Creates a new target for the boolean model class (classic subgroup discovery).
        If target_attribute and target_value are given, the target_selector is computed
        using attribute and value
        """
        if target_attribute is not None and target_value is not None:
            if target_selector is not None:
                raise ValueError(
                    "BinaryTarget is to be constructed"
                    "EITHER by a selector OR by attribute/value pair"
                )
            target_selector = EqualitySelector(target_attribute, target_value)
        if target_selector is None:
            raise ValueError("No target selector given")
        self.target_selector = target_selector

    def __repr__(self):
        return "T: " + str(self.target_selector)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __lt__(self, other):
        return str(self) < str(other)

    def covers(self, instance):
        return self.target_selector.covers(instance)

    def get_attributes(self):
        return (self.target_selector.attribute_name,)

    def get_base_statistics(self, subgroup, data):
        cover_arr, size_sg = get_cover_array_and_size(subgroup, len(data), data)
        positives = self.covers(data)
        instances_subgroup = size_sg
        positives_dataset = np.sum(positives)
        instances_dataset = len(data)
        positives_subgroup = np.sum(positives[cover_arr])
        return (
            instances_dataset,
            positives_dataset,
            instances_subgroup,
            positives_subgroup,
        )

    def calculate_statistics(self, subgroup, data, cached_statistics=None):
        if self.all_statistics_present(cached_statistics):
            return cached_statistics

        (
            instances_dataset,
            positives_dataset,
            instances_subgroup,
            positives_subgroup,
        ) = self.get_base_statistics(subgroup, data)
        statistics = {}
        statistics["size_sg"] = instances_subgroup
        statistics["size_dataset"] = instances_dataset
        statistics["positives_sg"] = positives_subgroup
        statistics["positives_dataset"] = positives_dataset
        statistics["size_complement"] = instances_dataset - instances_subgroup
        statistics["relative_size_sg"] = instances_subgroup / instances_dataset
        statistics["relative_size_complement"] = (
            instances_dataset - instances_subgroup
        ) / instances_dataset
        statistics["coverage_sg"] = positives_subgroup / positives_dataset
        statistics["coverage_complement"] = (
            positives_dataset - positives_subgroup
        ) / positives_dataset
        statistics["target_share_sg"] = positives_subgroup / instances_subgroup
        if instances_dataset == instances_subgroup:
            statistics["target_share_complement"] = float("nan")
        else:
            statistics["target_share_complement"] = (
                positives_dataset - positives_subgroup
            ) / (instances_dataset - instances_subgroup)
        statistics["target_share_dataset"] = positives_dataset / instances_dataset
        statistics["lift"] = (
            statistics["target_share_sg"] / statistics["target_share_dataset"]
        )
        return statistics


class SimplePositivesQF(
    AbstractInterestingnessMeasure
):  # pylint: disable=abstract-method
    tpl = namedtuple("PositivesQF_parameters", ("size_sg", "positives_count"))

    def __init__(self):
        self.dataset_statistics = None
        self.positives = None
        self.has_constant_statistics = False
        self.required_stat_attrs = ("size_sg", "positives_count")

    def calculate_constant_statistics(self, data, target):
        assert isinstance(target, BinaryTarget)
        self.positives = target.covers(data)
        self.dataset_statistics = SimplePositivesQF.tpl(
            len(data), np.sum(self.positives)
        )
        self.has_constant_statistics = True

    def calculate_statistics(
        self, subgroup, target, data, statistics=None
    ):  # pylint: disable=unused-argument
        cover_arr, size_sg = get_cover_array_and_size(
            subgroup, len(self.positives), data
        )
        return SimplePositivesQF.tpl(
            size_sg, np.count_nonzero(self.positives[cover_arr])
        )

    # <<< GpGrowth >>>
    def gp_get_stats(self, row_index):
        return np.array([1, self.positives[row_index]], dtype=int)

    def gp_get_null_vector(self):
        return np.zeros(2)

    def gp_merge(self, left, right):
        left += right

    def gp_get_params(self, _cover_arr, v):
        return SimplePositivesQF.tpl(v[0], v[1])

    def gp_to_str(self, stats):
        return " ".join(map(str, stats))

    def gp_size_sg(self, stats):
        return stats[0]

    @property
    def gp_requires_cover_arr(self):
        return False


# TODO Make ChiSquared useful for real nominal data not just binary
#      Introduce Enum for direction
#      Maybe it is possible to give a optimistic estimate for ChiSquared
class ChiSquaredQF(SimplePositivesQF):  # pragma: no cover
    """
    ChiSquaredQF which test for statistical independence
    of a subgroup against it's complement

    ...

    """

    @staticmethod
    def chi_squared_qf(
        instances_dataset,
        positives_dataset,
        instances_subgroup,
        positives_subgroup,
        min_instances=5,
        bidirect=True,
        direction_positive=True,
        index=0,
    ):
        """
        Performs chi2 test of statistical independence

        Test whether a subgroup is statistically independent
        from it's complement (see scipy.stats.chi2_contingency).


        Parameters
        ----------
        instances_dataset,
                positives_dataset,
                instances_subgroup,
                positives_subgroup : int
            counts of subgroup and dataset
        min_instances : int, optional
            number of required instances, if less -inf is returned for that subgroup
        bidirect : bool, optional
            If true both directions are considered interesting
            else direction_positive decides which direction is interesting
        direction_positive: bool, optional
            Only used if bidirect=False; specifies whether you are interested
            in positive (True) or negative deviations
        index : {0, 1}, optional
            decides whether the test statistic (0) or the p-value (1) should be used
        """
        import scipy.stats  # pylint:disable=import-outside-toplevel

        if (instances_subgroup < min_instances) or (
            (instances_dataset - instances_subgroup) < min_instances
        ):
            return float("-inf")

        negatives_subgroup = instances_subgroup - positives_subgroup
        negatives_dataset = instances_dataset - positives_dataset
        negatives_complement = negatives_dataset - negatives_subgroup
        positives_complement = positives_dataset - positives_subgroup

        val = scipy.stats.chi2_contingency(
            [
                [positives_subgroup, positives_complement],
                [negatives_subgroup, negatives_complement],
            ],
            correction=False,
        )[index]
        if bidirect:
            return val
        p_subgroup = positives_subgroup / instances_subgroup
        p_dataset = positives_dataset / instances_dataset
        if direction_positive and p_subgroup > p_dataset:
            return val
        if not direction_positive and p_subgroup < p_dataset:
            return val
        return -val

    @staticmethod
    def chi_squared_qf_weighted(
        subgroup,
        data,
        weighting_attribute,
        effective_sample_size=0,
        min_instances=5,
    ):
        import scipy.stats  # pylint:disable=import-outside-toplevel

        (
            instancesDataset,
            positivesDataset,
            instancesSubgroup,
            positivesSubgroup,
        ) = subgroup.get_base_statistics(data, weighting_attribute)
        if (instancesSubgroup < min_instances) or (
            (instancesDataset - instancesSubgroup) < 5
        ):
            return float("inf")
        if effective_sample_size == 0:
            effective_sample_size = derive_effective_sample_size(
                data[weighting_attribute]
            )
        # p_subgroup = positivesSubgroup / instancesSubgroup
        # p_dataset = positivesDataset / instancesDataset

        negatives_subgroup = instancesSubgroup - positivesSubgroup
        negatives_dataset = instancesDataset - positivesDataset
        positives_complement = positivesDataset - positivesSubgroup
        negatives_complement = negatives_dataset - negatives_subgroup
        val = scipy.stats.chi2_contingency(
            [
                [positivesSubgroup, positives_complement],
                [negatives_subgroup, negatives_complement],
            ],
            correction=True,
        )[0]
        return scipy.stats.chi2.sf(val * effective_sample_size / instancesDataset, 1)

    def __init__(self, direction="both", min_instances=5, stat="chi2"):
        """
        Parameters
        ----------
        direction : {'both', 'positive', 'negative'}
            direction of deviation that is of interest
        min_instances : int, optional
            number of required instances, if less -inf is returned for that subgroup
        stat : {'chi2', 'p'}
            whether to report the test statistic
            or the p-value (see scipy.stats.chi2_contingency)
        """
        if direction == "both":
            self.bidirect = True
            self.direction_positive = True
        if direction == "positive":
            self.bidirect = False
            self.direction_positive = True
        if direction == "negative":
            self.bidirect = False
            self.direction_positive = False
        self.min_instances = min_instances
        self.index = {"chi2": 0, "p": 1}[stat]
        super().__init__()

    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        dataset = self.dataset_statistics
        return ChiSquaredQF.chi_squared_qf(
            dataset.size_sg,
            dataset.positives_count,
            statistics.size_sg,
            statistics.positives_count,
            self.min_instances,
            self.bidirect,
            self.direction_positive,
            self.index,
        )


class StandardQF(SimplePositivesQF, BoundedInterestingnessMeasure):
    """
    StandardQF which weights the relative size against the difference in averages

    The StandardQF is a general form of quality function
    which for different values of a is order equivalen to
    many popular quality measures.

    Attributes
    ----------
    a : float
        used as an exponent to scale the relative size to the difference in averages

    """

    @staticmethod
    def standard_qf(
        a, instances_dataset, positives_dataset, instances_subgroup, positives_subgroup
    ):
        if not hasattr(instances_subgroup, "__array_interface__") and (
            instances_subgroup == 0
        ):
            return np.nan
        p_subgroup = np.divide(positives_subgroup, instances_subgroup)
        # if instances_subgroup == 0:
        #    return 0
        # p_subgroup = positives_subgroup / instances_subgroup
        p_dataset = positives_dataset / instances_dataset
        return (instances_subgroup / instances_dataset) ** a * (p_subgroup - p_dataset)

    def __init__(self, a):
        """
        Parameters
        ----------
        a : float
            exponent to trade-off the relative size with the difference in means
        """
        self.a = a
        super().__init__()

    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        dataset = self.dataset_statistics
        return StandardQF.standard_qf(
            self.a,
            dataset.size_sg,
            dataset.positives_count,
            statistics.size_sg,
            statistics.positives_count,
        )

    def optimistic_estimate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        dataset = self.dataset_statistics
        return StandardQF.standard_qf(
            self.a,
            dataset.size_sg,
            dataset.positives_count,
            statistics.positives_count,
            statistics.positives_count,
        )

    def optimistic_generalisation(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        dataset = self.dataset_statistics
        pos_remaining = dataset.positives_count - statistics.positives_count
        return StandardQF.standard_qf(
            self.a,
            dataset.size_sg,
            dataset.positives_count,
            statistics.size_sg + pos_remaining,
            dataset.positives_count,
        )


class LiftQF(StandardQF):
    """
    Lift Quality Function

    LiftQF is a StandardQF with a=0.
    Thus it treats the difference in ratios as the quality
    without caring about the relative size of a subgroup.
    """

    def __init__(self):
        """ """

        super().__init__(0.0)


# TODO add true binomial quality function as in
# https://opus.bibliothek.uni-wuerzburg.de/opus4-wuerzburg/frontdoor/index/index/docId/1786 # noqa: E501
class SimpleBinomialQF(StandardQF):
    """
    Simple Binomial Quality Function

    SimpleBinomialQF is a StandardQF with a=0.5.
    It is an order equivalent approximation of the full binomial test
    if the subgroup size is much smaller than the size of the entire dataset.
    """

    def __init__(self):
        """ """

        super().__init__(0.5)


class WRAccQF(StandardQF):
    """
    Weighted Relative Accuracy Quality Function

    WRAccQF is a StandardQF with a=1.
    It is order equivalent to the difference in the observed
    and expected number of positive instances.
    """

    def __init__(self):
        """ """

        super().__init__(1.0)


#####
# GeneralizationAware Interestingness Measures
#####
class GeneralizationAware_StandardQF(GeneralizationAwareQF_stats):
    def __init__(self, a):
        super().__init__(StandardQF(0))
        self.a = a

    def get_max(self, *args):
        max_ratio = 0.0
        max_stats = None
        for stat in args:
            assert stat.size_sg > 0
            ratio = stat.positives_count / stat.size_sg
            if ratio > max_ratio:
                max_ratio = ratio
                max_stats = stat
        return max_stats

    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        sg_stats = statistics.subgroup_stats
        general_stats = statistics.generalisation_stats
        if sg_stats.size_sg == 0 or general_stats.size_sg == 0:
            return np.nan

        sg_ratio = sg_stats.positives_count / sg_stats.size_sg
        general_ratio = general_stats.positives_count / general_stats.size_sg
        return (sg_stats.size_sg / self.stats0.size_sg) ** self.a * (
            sg_ratio - general_ratio
        )



import unittest

from t_utils import assertResultEqual

import pysubgroup as ps
from pysubgroup import model_target
from pysubgroup.datasets import get_credit_data


class TestGpGrowth(unittest.TestCase):
    def test_gp_growth(self):
        data = get_credit_data()
        # warnings.filterwarnings("error")
        searchSpace_Nominal = ps.create_nominal_selectors(
            data, ignore=["duration", "credit_amount", "class"]
        )
        searchSpace_Numeric = ps.create_numeric_selectors(
            data, ignore=["duration", "credit_amount", "class"]
        )
        searchSpace = searchSpace_Nominal + searchSpace_Numeric

        target = model_target.EMM_Likelihood(
            model_target.PolyRegression_ModelClass(
                x_name="duration", y_name="credit_amount"
            )
        )
        qf = target
        task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=30,
            depth=2,
            qf=qf,
            constraints=[ps.MinSupportConstraint(100)],
        )
        results = ps.GpGrowth().execute(task)

        assertResultEqual(
            self,
            results,
            """0.013797436813222048 property_magnitude=='b'car'' AND purpose=='b'radio/tv''
            0.009898734286313158 credit_history=='b'existing paid'' AND job=='b'unskilled resident''
            0.008246523912217767 employment=='b'>=7'' AND existing_credits==2.0
            0.008140633988502345 residence_since==1.0
            0.0077895968955903 installment_commitment==4.0 AND personal_status=='b'female div/dep/mar''
            0.0074121246910140584 employment=='b'4<=X<7'' AND personal_status=='b'male single''
            0.0071337090733226 employment=='b'<1'' AND existing_credits==1.0
            0.0068670509196103215 checking_status=='b'no checking'' AND residence_since==2.0
            0.006504784620278891 residence_since==2.0 AND savings_status=='b'<100''
            0.006155044522767603 own_telephone=='b'none'' AND property_magnitude=='b'real estate''
            0.005861872663834833 housing=='b'rent'' AND residence_since==4.0
            0.005860952266316298 housing=='b'own'' AND purpose=='b'furniture/equipment''
            0.005803210503193562 property_magnitude=='b'car'' AND residence_since==2.0
            0.005799115508272782 installment_commitment==4.0 AND purpose=='b'radio/tv''
            0.005784113955920163 checking_status=='b'no checking'' AND purpose=='b'radio/tv''
            0.00576020125931543 own_telephone=='b'yes'' AND personal_status=='b'female div/dep/mar''
            0.005710873446125693 age: [30.0:36.0[ AND own_telephone=='b'yes''
            0.005447977966354751 employment=='b'>=7'' AND personal_status=='b'male single''
            0.005369226054398001 credit_history=='b'critical/other existing credit'' AND own_telephone=='b'yes''
            0.0051795091578145415 checking_status=='b'<0''
            0.005142633556137799 installment_commitment==3.0 AND job=='b'skilled''
            0.005113508747833315 personal_status=='b'female div/dep/mar'' AND property_magnitude=='b'car''
            0.005076244770614837 job=='b'unskilled resident'' AND num_dependents==1.0
            0.005037829303302348 checking_status=='b'no checking'' AND employment=='b'>=7''
            0.004968598479332576 age: [26.0:30.0[ AND own_telephone=='b'none''
            0.004948240707577685 employment=='b'<1'' AND other_payment_plans=='b'none''
            0.004908520358522154 age<26.0 AND existing_credits==1.0
            0.00483781209051734 installment_commitment==3.0 AND other_payment_plans=='b'none''
            0.00476540157528357 own_telephone=='b'none'' AND purpose=='b'furniture/equipment''
            0.004700479771495154 employment=='b'1<=X<4'' AND existing_credits==2.0""",  # noqa: E501
        )


if __name__ == "__main__":
    unittest.main()



from collections import defaultdict


def assertResultEqual(self, result, s):
    alg_results = to_dict(result.to_descriptions())
    sol_results = to_dict(conjunctions_from_str(s))
    self.maxDiff = None
    self.assertEqual(len(alg_results), len(sol_results))
    # self.assertDictEqual(alg_results, sol_results)
    self.assertListEqual(sorted(alg_results.keys()), sorted(sol_results.keys()))
    for key in alg_results.keys():
        self.assertListEqual(alg_results[key], sol_results[key], msg=f"{key}")


def print_result(result):
    for _, (quality, pattern, _) in enumerate(
        result.to_descriptions(include_stats=True)
    ):
        print(quality, pattern)


def conjunctions_from_str(s):
    import pysubgroup as ps  # pylint:disable = import-outside-toplevel

    result = []
    for line in s.strip().splitlines():
        parts = line.strip().split(" ")
        q = float(parts[0])
        tmp = " ".join(parts[1:])
        result.append((q, ps.Conjunction.from_str(tmp)))
    return result


def interval_selectors_from_str(s):
    import pysubgroup as ps  # pylint:disable = import-outside-toplevel

    result = []
    for line in s.strip().splitlines():
        result.append(ps.IntervalSelector.from_str(line))
    return result


def to_dict(result):
    out = defaultdict(list)
    for tpl in result:
        q = round(float(tpl[0]), 8)
        sg = tpl[1]
        out[q].append(sg)
    return {key: sorted(out[key]) for key in sorted(out.keys())}



"""
Created on 29.09.2017

@author: lemmerfn
"""
from collections import namedtuple
from functools import total_ordering

import pysubgroup as ps


@total_ordering
class FITarget(ps.BaseTarget):
    statistic_types = ("size_sg", "size_dataset")

    def __repr__(self):
        return "T: Frequent Itemsets"

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __lt__(self, other):
        return str(self) < str(other)  # pragma: no cover

    def get_attributes(self):
        return []

    def get_base_statistics(self, subgroup, data):
        _, size = ps.get_cover_array_and_size(subgroup, len(data), data)
        return size

    def calculate_statistics(self, subgroup_description, data, cached_statistics=None):
        if self.all_statistics_present(cached_statistics):
            return cached_statistics

        _, size = ps.get_cover_array_and_size(subgroup_description, len(data), data)
        statistics = {}
        statistics["size_sg"] = size
        statistics["size_dataset"] = len(data)
        return statistics


class SimpleCountQF(ps.AbstractInterestingnessMeasure):
    tpl = namedtuple("CountQF_parameters", ("size_sg"))
    gp_requires_cover_arr = False

    def __init__(self):
        self.required_stat_attrs = ("size_sg",)
        self.has_constant_statistics = True
        self.size_dataset = None

    def calculate_constant_statistics(
        self, data, target
    ):  # pylint: disable=unused-argument
        self.size_dataset = len(data)

    def calculate_statistics(
        self, subgroup_description, target, data, statistics=None
    ):  # pylint: disable=unused-argument
        _, size = ps.get_cover_array_and_size(
            subgroup_description, self.size_dataset, data
        )
        return SimpleCountQF.tpl(size)

    def gp_get_stats(self, _):
        return {"size_sg": 1}

    def gp_get_null_vector(self):
        return {"size_sg": 0}

    def gp_merge(self, left, right):
        left["size_sg"] += right["size_sg"]

    def gp_get_params(self, _cover_arr, v):
        return SimpleCountQF.tpl(v["size_sg"])

    def gp_to_str(self, stats):
        return str(stats["size_sg"])

    def gp_size_sg(self, stats):
        return stats["size_sg"]


class CountQF(SimpleCountQF, ps.BoundedInterestingnessMeasure):
    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        return statistics.size_sg

    def optimistic_estimate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        return statistics.size_sg


class AreaQF(SimpleCountQF):
    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        return statistics.size_sg * subgroup.depth



import unittest

import matplotlib.pyplot as plt

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data, get_titanic_data

plt.interactive(False)


class TestVizualization(unittest.TestCase):
    def setUp(self):
        data = get_titanic_data()

        target = ps.BinaryTarget("Survived", True)
        searchspace = ps.create_selectors(data, ignore=["Survived"])
        task = ps.SubgroupDiscoveryTask(
            data, target, searchspace, result_set_size=5, depth=2, qf=ps.WRAccQF()
        )
        self.result = ps.DFS(ps.BitSetRepresentation).execute(task)
        self.result_df = self.result.to_dataframe()
        self.data = data

    def test_plot_sgbars(self):
        ps.plot_sgbars(
            self.result_df,
        )
        ps.plot_sgbars(self.result_df, dynamic_widths=True)

    def test_plot_roc(self):
        ps.plot_roc(self.result_df, self.data)
        ps.plot_roc(self.result_df, self.data, annotate=True)

    def test_plot_npspace(self):
        ps.plot_npspace(self.result_df, self.data)
        ps.plot_npspace(self.result_df, self.data, annotate=False)
        ps.plot_npspace(self.result_df, self.data, fixed_limits=True)

    def test_similarity_dendogram(self):
        ps.similarity_dendrogram(self.result, self.data)
        ps.similarity_dendrogram(self.result.to_descriptions(), self.data)

    def test_similarity_sgs(self):
        ps.similarity_sgs(self.result.to_descriptions(), self.data, color=False)
        try:
            ps.similarity_sgs(self.result.to_descriptions(), self.data, color=True)
        except ImportError:
            pass

    def test_supportSetVisualization(self):
        ps.supportSetVisualization(self.result)
        ps.supportSetVisualization(self.result, in_order=False)
        ps.supportSetVisualization(self.result, drop_empty=False)


class TestVizualization2(unittest.TestCase):
    def setUp(self):
        self.data = get_credit_data()
        self.target = ps.NumericTarget("credit_amount")
        searchSpace_Nominal = ps.create_nominal_selectors(
            self.data, ignore=["credit_amount"]
        )
        searchSpace_Numeric = (
            []
        )  # ps.create_numeric_selectors(data, ignore=['credit_amount'], nbins=10)
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        task = ps.SubgroupDiscoveryTask(
            self.data,
            self.target,
            searchSpace,
            result_set_size=10,
            depth=2,
            qf=ps.StandardQFNumeric(1, False, estimator="max", centroid="median"),
        )
        self.result = ps.DFS(ps.BitSetRepresentation).execute(task)

    def test_plot_distribution_numeric(self):
        ps.plot_distribution_numeric(self.result, self.target, self.data, 20)
        ps.plot_distribution_numeric(
            self.result, self.target, self.data, 20, show_dataset=False
        )
        ps.plot_distribution_numeric(
            self.result.to_descriptions(), self.target, self.data, 20
        )
        ps.plot_distribution_numeric(
            self.result.to_descriptions()[0][1], self.target, self.data, 20
        )

        subgroups = [subgroup for quality, subgroup in self.result.to_descriptions()]
        ps.plot_distribution_numeric(subgroups, self.target, self.data, 20)


if __name__ == "__main__":
    unittest.main()



# flake8: noqa: E501

import unittest

import pytest
from algorithms_testing import TestAlgorithmsBase

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestSettings:
    All = True
    Apriori = False
    SimpleDFS = False
    BestFirstSearch = False
    BeamSearch = False
    DFS_bitset = False
    DFS_set = False
    DFS_numpyset = False
    SimpleSearch = False


skip_long_running = True


class BooleanTargetBase(TestAlgorithmsBase):
    # pylint: disable=no-member
    @unittest.skipUnless(TestSettings.All or TestSettings.Apriori, "flag not set")
    def test_Apriori(self):
        self.runAlgorithm(
            ps.Apriori(), "Apriori", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.SimpleDFS, "flag not set")
    def test_SimpleDFS(self):
        self.runAlgorithm(
            ps.SimpleDFS(), "SimpleDFS", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(
        TestSettings.All or TestSettings.BestFirstSearch, "flag not set"
    )
    def test_BestFirstSearch(self):
        self.runAlgorithm(
            ps.BestFirstSearch(),
            "BestFirstSearch",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.BeamSearch, "flag not set")
    def test_BeamSearch(self):
        self.runAlgorithm(
            ps.BeamSearch(beam_width=self.task.result_set_size),
            "BeamSearch",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_bitset, "flag not set")
    def test_DFS_bitset(self):
        self.runAlgorithm(
            ps.DFS(ps.BitSetRepresentation),
            "DFS bitset",
            self.result,
            self.qualities,
            self.task,
        )

    @pytest.mark.slow
    @unittest.skipUnless(TestSettings.SimpleSearch, "flag not set")
    def test_SimpleSearch_slow(self):
        self.runAlgorithm(
            ps.SimpleSearch(), "SimpleSearch", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_set, "flag not set")
    def test_DFS_set(self):
        self.runAlgorithm(
            ps.DFS(ps.SetRepresentation),
            "DFS set",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_numpyset, "flag not set")
    def test_DFS_numpy_sets(self):
        self.runAlgorithm(
            ps.DFS(ps.NumpySetRepresentation),
            "DFS numpyset",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.GpGrowth, "flag not set")
    def test_gp_growth(self):
        self.runAlgorithm(
            ps.GpGrowth(), "GpGrowth", self.result, self.qualities, self.task
        )

    def evaluate_result(self, algorithm_result, result, qualities):
        df = algorithm_result.to_dataframe()
        self.assertTrue(
            all(algorithm_result.task.constraints[0].min_support <= df["size_sg"])
        )
        TestAlgorithmsBase.evaluate_result(self, algorithm_result, result, qualities)

    # pylint: enable=no-member


class TestAlgorithms(BooleanTargetBase, unittest.TestCase):
    def setUp(self):
        NS_checking = ps.EqualitySelector("checking_status", b"<0")
        NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
        NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        NS_payment_plans = ps.EqualitySelector("other_payment_plans", b"none")
        self.result = [
            ps.Conjunction([NS_checking, NS_foreign_worker]),
            ps.Conjunction([NS_checking]),
            ps.Conjunction([NS_checking, NS_other_parties, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_savings_status]),
            ps.Conjunction([NS_checking, NS_foreign_worker, NS_payment_plans]),
            ps.Conjunction([NS_checking, NS_payment_plans]),
            ps.Conjunction([NS_foreign_worker, NS_savings_status]),
            ps.Conjunction([NS_foreign_worker, NS_other_parties, NS_savings_status]),
        ]
        self.qualities = [
            0.055299999999999995,
            0.05280000000000001,
            0.052300000000000006,
            0.05059999999999999,
            0.04959999999999999,
            0.048299999999999996,
            0.0426,
            0.04,
            0.03869999999999999,
            0.03750000000000001,
        ]
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace = ps.create_nominal_selectors(data, ignore=["class"])
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=5,
            qf=ps.StandardQF(1.0),
            constraints=[ps.MinSupportConstraint(200)],
        )


# class TestAlgorithms2(BooleanTargetBase, unittest.TestCase):
#     def setUp(self):
#         NS_checking = ps.EqualitySelector("checking_status", b"<0")
#         NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
#         NS_other_parties = ps.EqualitySelector("other_parties", b"none")
#         NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
#         NS_payment_plans = ps.EqualitySelector("other_payment_plans", b"none")
#         NS_job = ps.EqualitySelector("job", b"skilled")
#         self.result = [
#             ps.Conjunction([NS_savings_status]),
#             ps.Conjunction([NS_foreign_worker]),
#             ps.Conjunction([NS_foreign_worker, NS_other_parties]),
#             ps.Conjunction([NS_foreign_worker, NS_job]),
#         ]
#         self.qualities = [
#             0.03610000000000002,
#             0.0071000000000000256,
#             0.004999999999999996,
#             0.0006999999999999932,
#         ]
#         data = get_credit_data()
#         target = ps.BinaryTarget("class", b"bad")
#         searchSpace = ps.create_nominal_selectors(data, ignore=["class"])
#         self.task = ps.SubgroupDiscoveryTask(
#             data,
#             target,
#             searchSpace,
#             result_set_size=10,
#             depth=5,
#             qf=ps.StandardQF(1.0),
#             constraints=[ps.MinSupportConstraint(600)],
#         )


#   0.055299999999999995:        checking_status=='b'<0'' AND foreign_worker=='b'yes''
#   0.05280000000000001: checking_status=='b'<0''
#   0.052300000000000006:        checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none''
#   0.05059999999999999: checking_status=='b'<0'' AND other_parties=='b'none''
#   0.04959999999999999: checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND savings_status=='b'<100''
#   0.048299999999999996:        checking_status=='b'<0'' AND savings_status=='b'<100''
#   0.0426:      checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_payment_plans=='b'none''
#   0.04:        checking_status=='b'<0'' AND other_payment_plans=='b'none''
#   0.03869999999999999: foreign_worker=='b'yes'' AND savings_status=='b'<100''
#   0.03750000000000001: foreign_worker=='b'yes'' AND other_parties=='b'none'' AND savings_status=='b'<100''


# also includes numeric attributes and has 12 targets


if __name__ == "__main__":
    # unittest.main()
    suites = []
    suites.append(unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms))
    # suites.append(unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms2))
    # suites.append(unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms2))
    # suites.append( unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms3))
    complete_suite = unittest.TestSuite(suites)
    unittest.TextTestRunner(verbosity=2).run(complete_suite)

    # import cProfile
    # p=cProfile.Profile()
    # p.enable()
    # t=TestAlgorithms()
    # t.setUp()
    # t.test_BSD()

    # p.disable()
    # p.dump_stats(r"E:\SGD\profile2.prof")

# 0.055299999999999995:   checking_status=b'<0' AND foreign_worker=b'yes'
# 0.05280000000000001:    checking_status=b'<0'
# 0.052300000000000006:   checking_status=b'<0' AND other_parties=b'none' AND foreign_worker=b'yes'
# 0.05059999999999999:    checking_status=b'<0' AND other_parties=b'none'
# 0.04959999999999999:    checking_status=b'<0' AND savings_status=b'<100' AND foreign_worker=b'yes'
# 0.048299999999999996:   checking_status=b'<0' AND savings_status=b'<100'
# 0.04660000000000001:    checking_status=b'<0' AND savings_status=b'<100' AND other_parties=b'none' AND foreign_worker=b'
# 0.04550000000000001:    checking_status=b'<0' AND job=b'skilled' AND foreign_worker=b'yes'
# 0.0452:                 checking_status=b'<0' AND savings_status=b'<100' AND other_parties=b'none'
# 0.044399999999999995:   checking_status=b'<0' AND job=b'skilled'

# <<< best 12 for a=1
# 0.055299999999999995: checking_status=='b'<0'' AND foreign_worker=='b'yes''
# 0.05280000000000001:  checking_status=='b'<0''
# 0.052300000000000006: checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none''
# 0.05059999999999999:  checking_status=='b'<0'' AND other_parties=='b'none''
# 0.04959999999999999:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND savings_status=='b'<100''
# 0.04870000000000001:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND num_dependents==1.0
# 0.048299999999999996: checking_status=='b'<0'' AND savings_status=='b'<100''
# 0.0474:               checking_status=='b'<0'' AND num_dependents==1.0
# 0.04660000000000001:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.04550000000000001:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled''


# <<< best 12 a=0.5
# 0.11457431093955019:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.113713540226172:    checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND savings_status=='b'<100''
# 0.11201325679119281:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled''
# 0.1117538749727658:   checking_status=='b'<0'' AND job=='b'skilled'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.11161046793076415:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND other_parties=='b'none''
# 0.11145710640046322:  checking_status=='b'<0'' AND job=='b'skilled'' AND savings_status=='b'<100''
# 0.11045259291161472:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.10929088624672183:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none''
# 0.10875519439407161:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND savings_status=='b'<100''
# 0.10866138825404954:  checking_status=='b'<0'' AND foreign_worker=='b'yes''
# 0.10832735026213287:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND num_dependents==1.0 AND savings_status=='b'<100''
# 0.10813405094128754:  checking_status=='b'<0'' AND job=='b'skilled'' AND other_parties=='b'none''



# flake8: noqa: E501

import unittest

import pytest
from algorithms_testing import TestAlgorithmsBase

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestSettings:
    All = True
    Apriori = False
    SimpleDFS = False
    BestFirstSearch = False
    BeamSearch = False
    DFS_bitset = False
    DFS_set = False
    DFS_numpyset = False
    SimpleSearch = False
    GpGrowth = False


skip_long_running = True


class BooleanTargetBase:
    # pylint: disable=no-member
    @unittest.skipUnless(TestSettings.All or TestSettings.Apriori, "flag not set")
    def test_Apriori(self):
        self.runAlgorithm(
            ps.Apriori(), "Apriori", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.Apriori, "flag not set")
    def test_Apriori_no_numba(self):
        self.runAlgorithm(
            ps.Apriori(use_numba=False),
            "Apriori",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.Apriori, "flag not set")
    def test_Apriori_no_vector(self):
        alg = ps.Apriori()
        alg.use_vectorization = False
        self.runAlgorithm(alg, "Apriori", self.result, self.qualities, self.task)

    @unittest.skipUnless(TestSettings.All or TestSettings.SimpleDFS, "flag not set")
    def test_SimpleDFS(self):
        self.runAlgorithm(
            ps.SimpleDFS(), "SimpleDFS", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(
        TestSettings.All or TestSettings.BestFirstSearch, "flag not set"
    )
    def test_BestFirstSearch(self):
        self.runAlgorithm(
            ps.BestFirstSearch(),
            "BestFirstSearch",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.BeamSearch, "flag not set")
    def test_BeamSearch(self):
        self.runAlgorithm(
            ps.BeamSearch(self.task.result_set_size),
            "BeamSearch",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_bitset, "flag not set")
    def test_DFS_bitset(self):
        self.runAlgorithm(
            ps.DFS(ps.BitSetRepresentation),
            "DFS bitset",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.GpGrowth, "flag not set")
    def test_gp_growth(self):
        self.task.constraints_monotone.append(
            ps.MinSupportConstraint(100)
        )  # constraint added to speed up tests
        self.runAlgorithm(
            ps.GpGrowth(), "GpGrowth", self.result, self.qualities, self.task
        )
        self.task.constraints_monotone.pop(-1)

    @unittest.skipUnless(TestSettings.All or TestSettings.GpGrowth, "flag not set")
    def test_gp_growth_long(self):
        self.runAlgorithm(
            ps.GpGrowth(), "GpGrowth", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.GpGrowth, "flag not set")
    def test_gp_growth_top_down(self):
        self.task.constraints_monotone.append(ps.MinSupportConstraint(100))
        self.runAlgorithm(
            ps.GpGrowth(mode="t_d"), "GpGrowth", self.result, self.qualities, self.task
        )
        self.task.constraints_monotone.pop(-1)

    @pytest.mark.slow
    @unittest.skipUnless(TestSettings.SimpleSearch, "flag not set")
    def test_SimpleSearch_slow(self):
        self.runAlgorithm(
            ps.SimpleSearch(), "SimpleSearch", self.result, self.qualities, self.task
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_set, "flag not set")
    def test_DFS_set(self):
        self.runAlgorithm(
            ps.DFS(ps.SetRepresentation),
            "DFS set",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_numpyset, "flag not set")
    def test_DFS_numpy_sets(self):
        self.runAlgorithm(
            ps.DFS(ps.NumpySetRepresentation),
            "DFS numpyset",
            self.result,
            self.qualities,
            self.task,
        )

    @unittest.skipUnless(TestSettings.All or TestSettings.DFS_numpyset, "flag not set")
    def test_DFS_default(self):
        self.runAlgorithm(
            ps.DFS(),
            "DFS default",
            self.result,
            self.qualities,
            self.task,
        )

    # pylint: enable=no-member


class TestAlgorithms(TestAlgorithmsBase, BooleanTargetBase, unittest.TestCase):
    def setUp(self):
        NS_checking = ps.EqualitySelector("checking_status", b"<0")
        NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
        NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        NS_job = ps.EqualitySelector("job", b"skilled")
        self.result = [
            ps.Conjunction([NS_checking, NS_foreign_worker]),
            ps.Conjunction([NS_checking]),
            ps.Conjunction([NS_checking, NS_other_parties, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_savings_status]),
            ps.Conjunction(
                [NS_checking, NS_savings_status, NS_other_parties, NS_foreign_worker]
            ),
            ps.Conjunction([NS_checking, NS_job, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_job]),
        ]
        self.qualities = [
            0.055299999999999995,
            0.05280000000000001,
            0.052300000000000006,
            0.05059999999999999,
            0.04959999999999999,
            0.048299999999999996,
            0.04660000000000001,
            0.04550000000000001,
            0.0452,
            0.044399999999999995,
        ]
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace = ps.create_nominal_selectors(data, ignore=["class"])
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=5,
            qf=ps.StandardQF(1.0),
        )


# also includes numeric attributes and has 12 targets


class TestAlgorithms2(TestAlgorithmsBase, BooleanTargetBase, unittest.TestCase):
    def setUp(self):
        NS_checking = ps.EqualitySelector("checking_status", b"<0")
        NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
        NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        NS_job = ps.EqualitySelector("job", b"skilled")
        NS_dependents = ps.EqualitySelector("num_dependents", 1.0)
        self.result = [
            ps.Conjunction([NS_checking, NS_foreign_worker]),
            ps.Conjunction([NS_checking]),
            ps.Conjunction([NS_checking, NS_other_parties, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_foreign_worker, NS_dependents]),
            ps.Conjunction([NS_checking, NS_savings_status]),
            ps.Conjunction([NS_checking, NS_dependents]),
            ps.Conjunction(
                [NS_checking, NS_savings_status, NS_other_parties, NS_foreign_worker]
            ),
            ps.Conjunction([NS_checking, NS_job, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_job]),
        ]
        self.qualities = [
            0.055299999999999995,
            0.05280000000000001,
            0.052300000000000006,
            0.05059999999999999,
            0.04959999999999999,
            0.04870000000000001,
            0.048299999999999996,
            0.0474,
            0.04660000000000001,
            0.04550000000000001,
            0.0452,
            0.044399999999999995,
        ]
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace_Nominal = ps.create_nominal_selectors(data, ignore=["class"])
        searchSpace_Numeric = ps.create_numeric_selectors(data, ignore=["class"])
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=12,
            depth=5,
            qf=ps.StandardQF(1.0),
        )

    def test_DFS_numpy_sets(self):
        super().test_DFS_numpy_sets()


# uses an a=0.5 and result_set_size = 12, is much faster because of that


class TestAlgorithms3(TestAlgorithmsBase, BooleanTargetBase, unittest.TestCase):
    def setUp(self):
        NS_checking = ps.EqualitySelector("checking_status", b"<0")
        NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
        NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        NS_job = ps.EqualitySelector("job", b"skilled")
        NS_dependents = ps.EqualitySelector("num_dependents", 1.0)
        self.result = [
            ps.Conjunction(
                [
                    NS_checking,
                    NS_foreign_worker,
                    NS_job,
                    NS_other_parties,
                    NS_savings_status,
                ]
            ),  # AND job=='b'skilled'' AND other_parties=='b'none'' AND savings_status=='b'<100'
            # 0.113713540226172:    checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND savings_status=='b'<100''
            ps.Conjunction([NS_checking, NS_foreign_worker, NS_job, NS_savings_status]),
            ps.Conjunction(
                [NS_checking, NS_foreign_worker, NS_job]
            ),  # checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled''
            # checking_status=='b'<0'' AND job=='b'skilled'' AND other_parties=='b'none'' AND savings_status=='b'<100''
            ps.Conjunction([NS_checking, NS_job, NS_other_parties, NS_savings_status]),
            ps.Conjunction([NS_checking, NS_foreign_worker, NS_job, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_job, NS_savings_status]),
            ps.Conjunction(
                [NS_checking, NS_foreign_worker, NS_other_parties, NS_savings_status]
            ),
            ps.Conjunction([NS_checking, NS_foreign_worker, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_foreign_worker, NS_savings_status]),
            ps.Conjunction([NS_checking, NS_foreign_worker]),
            ps.Conjunction(
                [
                    NS_checking,
                    NS_foreign_worker,
                    NS_job,
                    NS_dependents,
                    NS_savings_status,
                ]
            ),
            ps.Conjunction([NS_checking, NS_job, NS_other_parties]),
        ]

        self.qualities = [
            0.11457431093955019,
            0.113713540226172,
            0.11201325679119281,
            0.1117538749727658,
            0.11161046793076415,
            0.11145710640046322,
            0.11045259291161472,
            0.10929088624672183,
            0.10875519439407161,
            0.10866138825404954,
            0.10832735026213287,
            0.10813405094128754,
        ]
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace_Nominal = ps.create_nominal_selectors(data, ignore=["class"])
        searchSpace_Numeric = ps.create_numeric_selectors(data, ignore=["class"])
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=12,
            depth=5,
            qf=ps.StandardQF(0.5),
        )

    # @pytest.mark.slow  # not really that slow (< 1 min)
    def test_BestFirstSearch(self):
        super().test_BestFirstSearch()

    @pytest.mark.slow  #  (< 2 min)
    def test_SimpleDFS_slow(self):
        super().test_SimpleDFS()

    @pytest.mark.slow  # (< 1 min)
    def test_DFS_numpy_sets_slow(self):
        super().test_DFS_numpy_sets()


if __name__ == "__main__":
    # unittest.main()
    suites = []
    suites.append(unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms))
    suites.append(unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms2))
    # suites.append(unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms3))
    complete_suite = unittest.TestSuite(suites)
    unittest.TextTestRunner(verbosity=2).run(complete_suite)

    # import cProfile
    # p=cProfile.Profile()
    # p.enable()
    # t=TestAlgorithms()
    # t.setUp()
    # t.test_BSD()

    # p.disable()
    # p.dump_stats(r"E:\SGD\profile2.prof")

# 0.055299999999999995:   checking_status=b'<0' AND foreign_worker=b'yes'
# 0.05280000000000001:    checking_status=b'<0'
# 0.052300000000000006:   checking_status=b'<0' AND other_parties=b'none' AND foreign_worker=b'yes'
# 0.05059999999999999:    checking_status=b'<0' AND other_parties=b'none'
# 0.04959999999999999:    checking_status=b'<0' AND savings_status=b'<100' AND foreign_worker=b'yes'
# 0.048299999999999996:   checking_status=b'<0' AND savings_status=b'<100'
# 0.04660000000000001:    checking_status=b'<0' AND savings_status=b'<100' AND other_parties=b'none' AND foreign_worker=b'
# 0.04550000000000001:    checking_status=b'<0' AND job=b'skilled' AND foreign_worker=b'yes'
# 0.0452:                 checking_status=b'<0' AND savings_status=b'<100' AND other_parties=b'none'
# 0.044399999999999995:   checking_status=b'<0' AND job=b'skilled'

# <<< best 12 for a=1
# 0.055299999999999995: checking_status=='b'<0'' AND foreign_worker=='b'yes''
# 0.05280000000000001:  checking_status=='b'<0''
# 0.052300000000000006: checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none''
# 0.05059999999999999:  checking_status=='b'<0'' AND other_parties=='b'none''
# 0.04959999999999999:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND savings_status=='b'<100''
# 0.04870000000000001:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND num_dependents==1.0
# 0.048299999999999996: checking_status=='b'<0'' AND savings_status=='b'<100''
# 0.0474:               checking_status=='b'<0'' AND num_dependents==1.0
# 0.04660000000000001:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.04550000000000001:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled''


# <<< best 12 a=0.5
# 0.11457431093955019:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.113713540226172:    checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND savings_status=='b'<100''
# 0.11201325679119281:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled''
# 0.1117538749727658:   checking_status=='b'<0'' AND job=='b'skilled'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.11161046793076415:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND other_parties=='b'none''
# 0.11145710640046322:  checking_status=='b'<0'' AND job=='b'skilled'' AND savings_status=='b'<100''
# 0.11045259291161472:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none'' AND savings_status=='b'<100''
# 0.10929088624672183:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND other_parties=='b'none''
# 0.10875519439407161:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND savings_status=='b'<100''
# 0.10866138825404954:  checking_status=='b'<0'' AND foreign_worker=='b'yes''
# 0.10832735026213287:  checking_status=='b'<0'' AND foreign_worker=='b'yes'' AND job=='b'skilled'' AND num_dependents==1.0 AND savings_status=='b'<100''
# 0.10813405094128754:  checking_status=='b'<0'' AND job=='b'skilled'' AND other_parties=='b'none''



import pysubgroup as ps


class MinSupportConstraint:
    def __init__(self, min_support):
        self.min_support = min_support

    @property
    def is_monotone(self):
        return True

    def is_satisfied(self, subgroup, statistics=None, data=None):
        if hasattr(statistics, "size_sg"):
            return statistics.size_sg >= self.min_support
        if isinstance(statistics, dict) and "size_sg" in statistics:
            return statistics["size_sg"] >= self.min_support
        try:
            return ps.get_size(subgroup, len(data), data) >= self.min_support
        except AttributeError:  # special case for gp_growth
            return self.get_size_sg(statistics)

    def gp_prepare(self, qf):
        self.get_size_sg = (
            qf.gp_size_sg
        )  # pylint: disable=attribute-defined-outside-init

    def gp_is_satisfied(self, node):
        return self.get_size_sg(node) >= self.min_support



import unittest

from algorithms_testing import TestAlgorithmsBase

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data

skip_long_running = True


class BooleanTargetBase(TestAlgorithmsBase):
    # pylint: disable=no-member
    @unittest.skipIf(skip_long_running, "as skip_long_running flag is True")
    def test_GeneralisingBFS(self):
        self.runAlgorithm(
            ps.GeneralisingBFS(),
            "GeneralisingBFS",
            self.result,
            self.qualities,
            self.task,
        )

    def test_GeneralisingApriori(self):
        algorithm = ps.Apriori(combination_name="Disjunction")
        algorithm.optimistic_estimate_name = "optimistic_generalisation"
        self.runAlgorithm(
            algorithm, "GeneralisingApriori", self.result, self.qualities, self.task
        )

    # pylint: enable=no-member


class TestAlgorithms(BooleanTargetBase, unittest.TestCase):
    def setUp(self):
        NS_checking = ps.EqualitySelector("checking_status", b"<0")
        NS_checking2 = ps.EqualitySelector("checking_status", b"0<=X<200")
        NS_other_parties = ps.EqualitySelector("other_parties", b"co applicant")
        NS_other = ps.EqualitySelector("purpose", b"other")
        NS_repairs = ps.EqualitySelector("purpose", b"repairs")
        NS_purpose = ps.EqualitySelector("purpose", b"business")

        NS_history = ps.EqualitySelector("credit_history", b"no credits/all paid")
        NS_history2 = ps.EqualitySelector("credit_history", b"all paid")
        NS_empl = ps.EqualitySelector("employment", b"unemployed")
        NS_job = ps.EqualitySelector("job", b"unemp/unskilled non res")
        NS_bank = ps.EqualitySelector("other_payment_plans", b"bank")
        self.result = [
            ps.Disjunction([NS_checking, NS_checking2, NS_bank]),
            ps.Disjunction([NS_checking, NS_checking2, NS_history]),
            ps.Disjunction([NS_checking, NS_checking2]),
            ps.Disjunction([NS_checking, NS_checking2, NS_other]),
            ps.Disjunction([NS_checking, NS_checking2, NS_repairs]),
            ps.Disjunction([NS_checking, NS_checking2, NS_empl]),
            ps.Disjunction([NS_checking, NS_checking2, NS_other_parties]),
            ps.Disjunction([NS_checking, NS_checking2, NS_history2]),
            ps.Disjunction([NS_checking, NS_checking2, NS_purpose]),
            ps.Disjunction([NS_checking, NS_checking2, NS_job]),
        ]
        self.qualities = [
            0.0779,
            0.07740000000000002,
            0.0771,
            0.07680000000000001,
            0.07670000000000002,
            0.0767,
            0.07660000000000003,
            0.07650000000000003,
            0.07650000000000001,
            0.07600000000000001,
        ]
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace = ps.create_nominal_selectors(data, ignore=["class"])
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=3,
            qf=ps.StandardQF(1.0),
        )


if __name__ == "__main__":
    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms)
    complete_suite = unittest.TestSuite([suite1])
    unittest.TextTestRunner(verbosity=2).run(complete_suite)

# flake8: noqa: E501

# depth = 5, a=1.0
#   0.07800000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_parties=='b'co applicant'' OR other_payment_plans=='b'bank''
#   0.07800000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_parties=='b'co applicant'' OR other_payment_plans=='b'bank'' OR purpose=='b'other''
#   0.07790000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'no credits/all paid'' OR other_parties=='b'co applicant'' OR other_payment_plans=='b'bank''
#   0.0779:              checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_payment_plans=='b'bank''
#   0.0779:              checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_payment_plans=='b'bank'' OR purpose=='b'other''
#   0.07780000000000002: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'no credits/all paid'' OR other_payment_plans=='b'bank''
#   0.07780000000000002: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'no credits/all paid'' OR other_payment_plans=='b'bank'' OR purpose=='b'other''
#   0.07770000000000002: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'no credits/all paid'' OR other_payment_plans=='b'bank'' OR purpose=='b'repairs''
#   0.0776:              checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_parties=='b'co applicant'' OR other_payment_plans=='b'bank'' OR purpose=='b'repairs''
#   0.07750000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_payment_plans=='b'bank'' OR purpose=='b'repairs''
#   0.07750000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'all paid'' OR other_parties=='b'co applicant'' OR other_payment_plans=='b'bank''

# depth = 3, a=1.0
#  0.0779:              checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_payment_plans=='b'bank''
#  0.07740000000000002: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'no credits/all paid''
#  0.0771:              checking_status=='b'0<=X<200'' OR checking_status=='b'<0''
#  0.07680000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR purpose=='b'other''
#  0.07670000000000002: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR purpose=='b'repairs''
#  0.0767:              checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR employment=='b'unemployed''
#  0.07660000000000003: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR other_parties=='b'co applicant''
#  0.07650000000000003: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR credit_history=='b'all paid''
#  0.07650000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR purpose=='b'business''
#  0.07600000000000001: checking_status=='b'0<=X<200'' OR checking_status=='b'<0'' OR job=='b'unemp/unskilled non res''



def test_algorithms_beamsearch_optimal_result():
    import numpy as np
    import pandas as pd

    import pysubgroup as ps

    data = pd.DataFrame(
        np.array(
            [
                #    a b c d   target
                [1, 1, 0, 0, 1],
                [1, 1, 0, 0, 1],
                [1, 1, 0, 0, 0],
                [1, 1, 0, 0, 0],
                [0, 0, 1, 1, 1],
                [0, 0, 1, 1, 1],
                [0, 0, 1, 1, 1],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
            ]
        ),
        columns=["a", "b", "c", "d", "target"],
    )

    target = ps.BinaryTarget("target", 1)

    searchspace = [
        ps.EqualitySelector("a", 1),
        ps.EqualitySelector("b", 1),
        ps.EqualitySelector("c", 1),
        ps.EqualitySelector("d", 1),
    ]
    task = ps.SubgroupDiscoveryTask(
        data, target, searchspace, result_set_size=2, depth=2, qf=ps.StandardQF(a=0.5)
    )
    result = ps.BeamSearch(beam_width=3).execute(task)

    assert result.results[0][1] == ps.Conjunction([searchspace[2], searchspace[3]])


def test_algorithms_beamsearch_width_too_small():
    import numpy as np
    import pandas as pd

    import pysubgroup as ps

    data = pd.DataFrame(
        np.array(
            [
                #    a b c d   target
                [1, 1, 0, 0, 1],
                [1, 1, 0, 0, 1],
                [1, 1, 0, 0, 0],
                [1, 1, 0, 0, 0],
                [0, 0, 1, 1, 1],
                [0, 0, 1, 1, 1],
                [0, 0, 1, 1, 1],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
            ]
        ),
        columns=["a", "b", "c", "d", "target"],
    )

    target = ps.BinaryTarget("target", 1)

    searchspace = [
        ps.EqualitySelector("a", 1),
        ps.EqualitySelector("b", 1),
        ps.EqualitySelector("c", 1),
        ps.EqualitySelector("d", 1),
    ]
    task = ps.SubgroupDiscoveryTask(
        data, target, searchspace, result_set_size=2, depth=2, qf=ps.StandardQF(a=0.5)
    )
    result = ps.BeamSearch(beam_width=2).execute(task)

    assert set([ps.Conjunction(s) for s in searchspace[:2]]) == set(
        [sg for _, sg, _ in result.results]
    )



import time
from collections import defaultdict

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data

data = get_credit_data()
# warnings.filterwarnings("error")
searchSpace_Nominal = ps.create_nominal_selectors(
    data, ignore=["duration", "credit_amount", "class"]
)
searchSpace_Numeric = ps.create_numeric_selectors(
    data, ignore=["duration", "credit_amount", "class"]
)
searchSpace = searchSpace_Nominal + searchSpace_Numeric
# searchSpace = [
#     ps.EqualitySelector("checking_status", b'<0'),
#     ps.EqualitySelector("own_telephone", b'yes'),
#     ps.IntervalSelector("age", 30, 36)]
# QF=model_target.EMM_Likelihood(
#     model_target.PolyRegression_ModelClass(x_name='duration', y_name='credit_amount'))
# target = ps.FITarget()
# QF=ps.CountQF()

QF = ps.StandardQF(0.5)
target = ps.BinaryTarget("class", b"bad")
constr = [ps.MinSupportConstraint(30)]
task = ps.SubgroupDiscoveryTask(
    data, target, searchSpace, result_set_size=30, depth=2, qf=QF, constraints=constr
)
ps.GpGrowth(mode="t_d").to_file(task, "E:/tmp/gp_credit.txt")


start_time = time.time()
gp = ps.GpGrowth(mode="t_d").execute(task)
print("--- %s seconds ---" % (time.time() - start_time))
# gp = [(qual, sg) for qual, sg in gp if sg.depth <= task.depth]
gp = sorted(gp.to_descriptions(include_stats=True), reverse=True)


def print_result(result):
    for i, (q, pattern, stats) in enumerate(gp):
        print(pattern, stats)
        if i > 10:
            break


start_time = time.time()
dfs1 = ps.SimpleDFS().execute(task)
print("--- %s seconds ---" % (time.time() - start_time))
dfs = dfs1.to_descriptions(include_stats=True)
dfs = sorted(dfs, reverse=True)
gp = sorted(gp, reverse=True)
# if len(gp[0][1].selectors)==0:
#    gp = gp[1:]


def better_sorted(tuples):
    the_dict = defaultdict(list)
    prev_key = tuples[0][0]
    for key, val in tuples:
        if abs(prev_key - key) < 10**-11:
            the_dict[prev_key].append(val)
        else:
            the_dict[key].append(val)
            prev_key = key
    print(len(the_dict))
    result = []
    for key, vals in the_dict.items():
        for val in sorted(vals):
            result.append((key, val))
    return result


# dfs = better_sorted(dfs)
# gp = better_sorted(gp)
# gp = gp[:task.result_set_size]
assert len(gp) == len(dfs), f"{len(gp)} != {len(dfs)}"
for i, (l, r) in enumerate(zip(gp, dfs)):
    print(i)
    print("gp:", l)
    print("df:", r)
    if not abs(l[0] - r[0]) < 10**-7:
        print("<<< ERROR <<<")
    # assert(abs(l[0]-r[0]) < 10 ** -7)
    # assert(l[1] == r[1])
    if i > 20:
        break
for q, sg, qual in dfs:
    print(str(q) + " " + str(sg))



import unittest

import numpy as np
import pandas as pd
from algorithms_testing import TestAlgorithmsBase

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestGeneralisationAwareQf(unittest.TestCase):
    def setUp(self):
        self.qf = ps.CountQF()
        self.ga_qf = ps.GeneralizationAwareQF(self.qf)
        self.prepare_df()

    def prepare_df(self):
        A = np.array([0, 0, 1, 1, 0, 0, 1, 1, 1, 1], dtype=bool)
        self.A1 = ps.EqualitySelector("columnA", True)
        self.A0 = ps.EqualitySelector("columnA", False)

        B = np.array(["A", "B", "C", "C", "B", "A", "D", "A", "A", "A"])
        self.BA = ps.EqualitySelector("columnB", "A")
        self.BC = ps.EqualitySelector("columnB", "C")
        self.BD = ps.EqualitySelector("columnB", "D")
        self.df = pd.DataFrame.from_dict(
            {
                "columnA": A,
                "columnB": B,
                "columnC": np.array([[0, 1] for _ in range(5)]).flatten(),
            }
        )

    def test_CountTarget1(self):
        df = self.df
        target = ps.FITarget()
        self.ga_qf.calculate_constant_statistics(df, target)

        ga_score = self.ga_qf.evaluate(ps.Conjunction([self.A1]), target, df)

        A1_score = self.qf.evaluate(ps.Conjunction([self.A1]), target, df)
        zero_score = self.qf.evaluate(ps.Conjunction([]), target, df)

        self.assertEqual(ga_score, A1_score - zero_score)

        ga2_score = self.ga_qf.evaluate(ps.Conjunction([self.A1]), target, df)

        self.assertEqual(ga2_score, ga_score)

    def test_CountTarget2(self):
        df = self.df
        target = ps.FITarget()
        self.ga_qf.calculate_constant_statistics(df, None)

        ga_score = self.ga_qf.evaluate(ps.Conjunction([self.A1, self.BA]), target, df)

        A_B_score = self.qf.evaluate(ps.Conjunction([self.A1, self.BA]), target, df)
        zero_score = self.qf.evaluate(ps.Conjunction([]), target, df)

        self.assertEqual(ga_score, A_B_score - zero_score)


class TestGeneralisationAware_StandardQf(unittest.TestCase):
    def setUp(self):
        self.df = None
        self.A1 = None
        self.BA = None
        self.A0 = None
        self.BD = None
        TestGeneralisationAwareQf.prepare_df(self)
        self.ga_qf = ps.GeneralizationAware_StandardQF(0)

    def test_simple(self):
        target = ps.BinaryTarget("columnC", 1)
        qf = ps.StandardQF(0)
        qf.calculate_constant_statistics(self.df, target)

        self.ga_qf.calculate_constant_statistics(self.df, target)

        # print(qf.calculate_statistics(self.A1, self.df))
        # print(qf.calculate_statistics(self.BA, self.df))
        # print(qf.calculate_statistics(ps.Conjunction([self.A1, self.BA]), self.df))
        # print(qf.calculate_statistics(slice(None), self.df))
        ga_stat = self.ga_qf.calculate_statistics(
            ps.Conjunction([self.A1, self.BA]), target, self.df
        )

        self.assertEqual(ga_stat.subgroup_stats, ps.SimplePositivesQF.tpl(3, 2))
        self.assertEqual(ga_stat.generalisation_stats, ps.SimplePositivesQF.tpl(5, 3))

        # Ensure cache works properly
        self.assertEqual(
            ga_stat,
            self.ga_qf.calculate_statistics(
                ps.Conjunction([self.A1, self.BA]), target, self.df
            ),
        )

        ga_score = self.ga_qf.evaluate(
            ps.Conjunction([self.A1, self.BA]), target, self.df
        )
        ga_score2 = self.ga_qf.evaluate(
            ps.Conjunction([self.A1, self.BA]), target, self.df
        )
        ga_score3 = self.ga_qf.evaluate(
            ps.Conjunction([self.A0, self.BD]), target, self.df
        )

        self.assertEqual(ga_score, ga_score2)
        self.assertAlmostEqual(ga_score, 0.06666666666666)
        self.assertTrue(np.isnan(ga_score3))


class TestAlgorithms(TestAlgorithmsBase, unittest.TestCase):
    # TODO properly specify desired result
    def setUp(self):
        NS_checking = ps.EqualitySelector("checking_status", b"<0")
        NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
        NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        NS_job = ps.EqualitySelector("job", b"skilled")
        self.result = [
            ps.Conjunction([NS_checking, NS_foreign_worker]),
            ps.Conjunction([NS_checking]),
            ps.Conjunction([NS_checking, NS_other_parties, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_savings_status]),
            ps.Conjunction(
                [NS_checking, NS_savings_status, NS_other_parties, NS_foreign_worker]
            ),
            ps.Conjunction([NS_checking, NS_job, NS_foreign_worker]),
            ps.Conjunction([NS_checking, NS_savings_status, NS_other_parties]),
            ps.Conjunction([NS_checking, NS_job]),
        ]
        self.qualities = [
            0.055299999999999995,
            0.05280000000000001,
            0.052300000000000006,
            0.05059999999999999,
            0.04959999999999999,
            0.048299999999999996,
            0.04660000000000001,
            0.04550000000000001,
            0.0452,
            0.044399999999999995,
        ]
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace = ps.create_nominal_selectors(data, ignore=["class"])
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=3,
            qf=ps.GeneralizationAwareQF(ps.StandardQF(1.0)),
        )

    @unittest.skip
    def test_GA_SimpleDFS(self):
        self.runAlgorithm(
            ps.SimpleDFS(), "SimpleDFS", self.result, self.qualities, self.task
        )

    @unittest.skip
    def test_StandardQF_GA_SimpleDFS(self):
        self.task.qf = ps.GeneralizationAware_StandardQF(0.5)
        self.runAlgorithm(
            ps.SimpleDFS(), "Standard_SimpleDFS", self.result, self.qualities, self.task
        )
        print(self.task.qf.cache)


if __name__ == "__main__":
    unittest.main()



import unittest

import pysubgroup as ps


class TestGeneralisationAwareQf(unittest.TestCase):
    def test_is_satisfied_dict(self):
        constr = ps.MinSupportConstraint(10)
        self.assertTrue(constr.is_satisfied(None, {"size_sg": 12}, None))
        self.assertFalse(constr.is_satisfied(None, {"size_sg": 9}, None))


if __name__ == "__main__":
    unittest.main()



import unittest

from algorithms_testing import TestAlgorithmsBase

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data

show_plots = False


class TestNominalTarget_to_result(unittest.TestCase, TestAlgorithmsBase):
    @classmethod
    def setUpClass(cls):
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace = ps.create_nominal_selectors(data, ignore=["class"])
        cls.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=5,
            qf=ps.StandardQF(1.0),
        )
        cls.result = ps.SimpleDFS().execute(cls.task)

    def test_to_dataframe(self):
        df = self.__class__.result.to_dataframe()
        self.assertEqual(len(df), 10)
        self.assertEqual(len(df.columns), 15)

    def test_to_descriptions(self):
        descriptions = self.__class__.result.to_descriptions()
        self.assertEqual(len(descriptions), 10)
        for tpl in descriptions:
            self.assertEqual(len(tpl), 2)
            self.assertTrue(isinstance(tpl[0], float))
            self.assertTrue(isinstance(tpl[1], ps.Conjunction))

    # def test_supportSetVisualization(self):
    #    img = self.__class__.result.supportSetVisualization()
    #    self.assertEqual(img.shape, (10, 274))
    #    if show_plots:
    #        plt.matshow(img)
    #        plt.show()


class TestNumericTarget_to_result(unittest.TestCase, TestAlgorithmsBase):
    @classmethod
    def setUpClass(cls):
        data = get_credit_data()
        target = ps.NumericTarget("credit_amount")
        searchSpace = ps.create_nominal_selectors(data, ignore=["credit_amount"])
        cls.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=3,
            qf=ps.StandardQFNumeric(1.0),
        )
        cls.result = ps.SimpleDFS().execute(cls.task)

    def test_to_dataframe(self):
        df = self.__class__.result.to_dataframe()
        self.assertEqual(len(df), 10)
        self.assertEqual(len(df.columns), 16)

    def test_to_descriptions(self):
        descriptions = self.__class__.result.to_descriptions()
        self.assertEqual(len(descriptions), 10)
        for tpl in descriptions:
            self.assertEqual(len(tpl), 2)
            self.assertTrue(isinstance(tpl[0], float))
            self.assertTrue(isinstance(tpl[1], ps.Conjunction))

    # def test_supportSetVisualization(self):
    #    img = self.__class__.result.supportSetVisualization()
    #    self.assertEqual(img.shape, (10, 425))
    #    if show_plots:
    #        plt.matshow(img)
    #        plt.show()


if __name__ == "__main__":
    show_plots = False
    unittest.main()



import unittest

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestUtilsDataConversion(unittest.TestCase):
    def setUp(self) -> None:
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace_Nominal = ps.create_nominal_selectors(data, ignore=["class"])
        searchSpace_Numeric = ps.create_numeric_selectors(data, ignore=["class"])
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        self.task = ps.SubgroupDiscoveryTask(
            data, target, searchSpace, result_set_size=2, depth=1, qf=ps.StandardQF(0.5)
        )
        self.result = ps.BeamSearch(1, beam_width_adaptive=True).execute(self.task)

    def test_to_dataframe(self):
        self.result.to_dataframe()

    def test_to_dataframe_inlude_target(self):
        self.result.to_dataframe(include_target=True)


if __name__ == "__main__":
    unittest.main()



import pysubgroup as ps
from pysubgroup.datasets import get_titanic_data

data = get_titanic_data()
target = ps.BinaryTarget("Survived", True)

searchspace = ps.create_selectors(data, ignore=["Survived"])
task = ps.SubgroupDiscoveryTask(
    data, target, searchspace, result_set_size=5, depth=2, qf=ps.ChiSquaredQF()
)
result = ps.BeamSearch().execute(task)

print(result.to_dataframe())



from functools import partial

import numpy as np

import pysubgroup as ps


def plot_sgbars(
    result_df,
    *,
    ylabel="target share",
    title="Discovered Subgroups",
    dynamic_widths=False,
    _suffix="",
):
    from matplotlib import pyplot as plt  # pylint: disable=import-outside-toplevel

    shares_sg = result_df["target_share_sg"]
    shares_compl = result_df["target_share_complement"]
    sg_relative_sizes = result_df["relative_size_sg"]
    x = np.arange(len(result_df))

    base_width = 0.8
    if dynamic_widths:
        width_sg = 0.02 + base_width * sg_relative_sizes
        width_compl = base_width - width_sg
    else:
        width_sg = base_width / 2
        width_compl = base_width / 2

    fig, ax = plt.subplots()
    rects1 = ax.bar(x, shares_sg, width_sg, align="edge")
    rects2 = ax.bar(
        x + width_sg, shares_compl, width_compl, align="edge", color="#61b76f"
    )

    ax.set_ylabel(ylabel)
    ax.set_title(title)
    ax.set_xticks(x + base_width / 2)
    ax.set_xticklabels(result_df.index, rotation=90)

    ax.legend((rects1[0], rects2[0]), ("subgroup", "complement"))
    fig.set_size_inches(12, len(result_df))

    return fig


def plot_roc(result_df, data, qf=ps.StandardQF(0.5), levels=40, annotate=False):
    from matplotlib import pyplot as plt  # pylint: disable=import-outside-toplevel

    assert isinstance(qf, ps.StandardQF)

    instances_dataset = len(data)
    positives_dataset = np.max(result_df["positives_dataset"])
    negatives_dataset = instances_dataset - positives_dataset

    xlist = np.linspace(0.01, 0.99, 100)
    ylist = np.linspace(0.01, 0.99, 100)
    X, Y = np.meshgrid(xlist, ylist)
    f = np.vectorize(
        partial(ps.StandardQF.standard_qf, qf.a, instances_dataset, positives_dataset),
        otypes=[np.float64],
    )
    Z = f(X * negatives_dataset + Y * positives_dataset, Y * positives_dataset)
    max_val = np.max([np.max(Z), -np.min(Z)])

    fig, ax = plt.subplots()
    cm = plt.colormaps["bwr"]

    plt.contourf(X, Y, Z, levels, cmap=cm, vmin=-max_val, vmax=max_val)

    for i, sg in result_df.iterrows():
        rel_positives_sg = sg["positives_sg"] / positives_dataset
        rel_negatives_sg = (sg["size_sg"] - sg["positives_sg"]) / negatives_dataset
        ax.plot(rel_negatives_sg, rel_positives_sg, "o", color="black")
        if annotate:
            label_margin = 0.01
            ax.annotate(
                str(i),
                (rel_negatives_sg + label_margin, rel_positives_sg + label_margin),
            )

    # plt.colorbar(cp)
    plt.title("Discovered subgroups")
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")

    return fig


def plot_npspace(result_df, data, annotate=True, fixed_limits=False):
    from matplotlib import pyplot as plt  # pylint: disable=import-outside-toplevel

    fig, ax = plt.subplots()

    for i, sg in result_df.iterrows():
        target_share_sg = sg["target_share_sg"]
        size_sg = sg["size_sg"]
        ax.plot(size_sg, target_share_sg, "o", color="black")
        if annotate:
            ax.annotate(str(i), (size_sg + 5, target_share_sg + 0.001))

    if fixed_limits:
        plt.xlim((0, len(data)))
        plt.ylim((0, 1))

    plt.title("Discovered subgroups")
    plt.xlabel("Size of Subgroup")
    plt.ylabel("Target Share Subgroup")

    return fig


def plot_distribution_numeric(sg, target, data, bins, show_dataset=True):
    from matplotlib import pyplot as plt  # pylint: disable=import-outside-toplevel

    if isinstance(sg, (list, tuple)):
        if isinstance(sg[0], tuple):
            list_sgs = [subgroup for quality, subgroup in sg]
        else:
            list_sgs = sg
    elif isinstance(sg, ps.SubgroupDiscoveryResult):
        list_sgs = [subgroup for quality, subgroup in sg.to_descriptions()]
    else:
        list_sgs = [sg]
    fig, _ = plt.subplots()
    for sg in list_sgs:
        target_values_sg = data[sg.covers(data)][target.get_attributes()].values
        target_values_data = data[target.get_attributes()].values
        plt.hist(
            target_values_sg,
            bins,
            alpha=0.5,
            label=str(sg),
            density=True,
        )
    if show_dataset:
        plt.hist(
            target_values_data, bins, alpha=0.5, label="Overall Data", density=True
        )
    plt.legend(loc="upper right")
    return fig


def similarity_sgs(sgd_results, data, color=True):
    import pandas as pd  # pylint:disable=import-outside-toplevel

    sgs = [x[1] for x in sgd_results]
    # sgNames = [str(sg.subgroup_description) for sg in sgs]
    dists = [[ps.overlap(sg, sg2, data) for sg2 in sgs] for sg in sgs]
    dist_df = pd.DataFrame(dists)
    if color:
        dist_df = dist_df.style.background_gradient()
    return dist_df


def similarity_dendrogram(result, data):
    from matplotlib import pyplot as plt  # pylint: disable=import-outside-toplevel
    from scipy.cluster.hierarchy import (  # pylint: disable=import-outside-toplevel
        dendrogram,
        linkage,
    )
    from scipy.spatial.distance import (
        squareform,  # pylint: disable=import-outside-toplevel
    )

    if isinstance(result, ps.SubgroupDiscoveryResult):
        result = result.to_descriptions()

    fig, _ = plt.subplots()
    dist_df = similarity_sgs(result, data, color=False)
    mat = 1 - dist_df.values
    dists = squareform(mat)
    linkage_matrix = linkage(dists, "single")
    dendrogram(linkage_matrix, labels=dist_df.index)
    return fig


def supportSetVisualization(result, in_order=True, drop_empty=True):
    df = result.task.data
    n_items = len(result.task.data)
    n_SGDs = len(result.results)
    covs = np.zeros((n_items, n_SGDs), dtype=bool)
    for i, (_, r) in enumerate(result.to_descriptions()):
        covs[:, i] = r.covers(df)

    img_arr = covs.copy()

    sort_inds_x = np.argsort(np.sum(covs, axis=1))[::-1]
    img_arr = img_arr[sort_inds_x, :]
    if not in_order:
        sort_inds_y = np.argsort(np.sum(covs, axis=0))
        img_arr = img_arr[:, sort_inds_y]
    if drop_empty:
        keep_entities = np.sum(img_arr, axis=1) > 0
        print(
            f"Discarding {n_items - np.count_nonzero(keep_entities)} "
            "entities that are not covered"
        )
        img_arr = img_arr[keep_entities, :]
    return img_arr.T
